Package: seqwrap
Type: Package
Title: Item-by-Item Iterative Model Fitting
Version: 0.7.0
Authors@R: c(
    person("Daniel", "Hammarstr\u00f6m",
    email = "daniel.hammarstrom@inn.no",
    role = c("aut", "cre", "cph"),
    comment = c(ORCID = "0000-0001-8360-2100")),
    person("Chidimma", "Echebiri",
    email = "chidimma.echebiri@inn.no",
    role = c("ctb"),
    comment = c(ORCID = "0000-0003-3134-5186")))
Description: Models high-dimensional data, such as RNA-seq or proteomic data
    using an item-by-item strategy. The package contains functions to wrap
    high-dimensional data and iterate over them using established R packages
    for regression modelling (e.g., 'glmmTMB' or 'mgcv').
License: GPL-3
Encoding: UTF-8
LazyData: true
LazyDataCompression: xz
URL: https://github.com/trainome/seqwrap
BugReports: https://github.com/trainome/seqwrap/issues
Depends: R (>= 4.1)
Imports: cli, S7, tibble, pbapply, parallel, stats, DHARMa, broom.mixed
Suggests: testthat (>= 3.0.0), dplyr, knitr, purrr, quarto, glmmTMB,
        lme4, nlme, MASS, rmarkdown, gt, edgeR, tidyselect, ggplot2,
        cowplot, ggtext, R.rsp
Config/testthat/edition: 3
Collate: 'seqwrap-global.R' 'aaa-.R' 'data-helper.R' 'data.R'
        'generic-summaries.R' 'seqwrap.R' 'seqwrap_mtf.R' 'simcounts.R'
        'simcounts2.R'
VignetteBuilder: quarto, R.rsp
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-15 14:17:25 UTC; 706194
Author: Daniel Hammarström [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-8360-2100>),
  Chidimma Echebiri [ctb] (ORCID:
    <https://orcid.org/0000-0003-3134-5186>)
Maintainer: Daniel Hammarström <daniel.hammarstrom@inn.no>
Repository: CRAN
Date/Publication: 2026-05-19 10:00:02 UTC
Built: R 4.5.2; ; 2026-05-19 13:03:30 UTC; unix

Package: seqwrap
Type: Package
Title: Item-by-Item Iterative Model Fitting
Version: 0.7.0
Authors@R: c(
    person("Daniel", "Hammarstr\u00f6m",
    email = "daniel.hammarstrom@inn.no",
    role = c("aut", "cre", "cph"),
    comment = c(ORCID = "0000-0001-8360-2100")),
    person("Chidimma", "Echebiri",
    email = "chidimma.echebiri@inn.no",
    role = c("ctb"),
    comment = c(ORCID = "0000-0003-3134-5186")))
Description: Models high-dimensional data, such as RNA-seq or proteomic data
    using an item-by-item strategy. The package contains functions to wrap
    high-dimensional data and iterate over them using established R packages
    for regression modelling (e.g., 'glmmTMB' or 'mgcv').
License: GPL-3
Encoding: UTF-8
LazyData: true
LazyDataCompression: xz
URL: https://github.com/trainome/seqwrap
BugReports: https://github.com/trainome/seqwrap/issues
Depends: R (>= 4.1)
Imports: cli, S7, tibble, pbapply, parallel, stats, DHARMa, broom.mixed
Suggests: testthat (>= 3.0.0), dplyr, knitr, purrr, quarto, glmmTMB,
        lme4, nlme, MASS, rmarkdown, gt, edgeR, tidyselect, ggplot2,
        cowplot, ggtext, R.rsp
Config/testthat/edition: 3
Collate: 'seqwrap-global.R' 'aaa-.R' 'data-helper.R' 'data.R'
        'generic-summaries.R' 'seqwrap.R' 'seqwrap_mtf.R' 'simcounts.R'
        'simcounts2.R'
VignetteBuilder: quarto, R.rsp
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-15 14:17:25 UTC; 706194
Author: Daniel Hammarström [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-8360-2100>),
  Chidimma Echebiri [ctb] (ORCID:
    <https://orcid.org/0000-0003-3134-5186>)
Maintainer: Daniel Hammarström <daniel.hammarstrom@inn.no>
Repository: CRAN
Date/Publication: 2026-05-19 10:00:02 UTC
Built: R 4.6.0; ; 2026-05-19 11:20:24 UTC; unix

# Generated by roxygen2: do not edit by hand

export(generic_evaluation)
export(generic_summary)
export(seqwrap)
export(seqwrapResults)
export(seqwrap_compose)
export(seqwrap_summarise)
export(simcounts)
export(simcounts2)
export(swcontainer)
importFrom(stats,as.formula)
importFrom(stats,median)
importFrom(stats,model.matrix)
importFrom(stats,predict)
importFrom(stats,rnbinom)
importFrom(stats,rnorm)
importFrom(stats,rpois)
importFrom(stats,runif)
importFrom(stats,sigma)

# seqwrap - Item-By-item Iterative Model Fitting


The `seqwrap` R package allows you to model high-dimensional data using
an item-by-item strategy with a specific model engine. Using `seqwrap`,
you can fit, e.g., gene/transcript data from an RNA sequencing
experiment to a model specified with random effects or custom
distributions. This is possible because `seqwrap` efficiently iterates
over all items (e.g., genes) and fits the data to the same model
formulation using established R packages for regression modeling.

See the package Vignettes for examples.

## Installing

Install the released version of `seqwrap` from CRAN:

``` r
install.packages("seqwrap")
```

Or install the development version from GitHub:

``` r
# install.packages("pak")
pak::pkg_install("trainome/seqwrap")
```

## -----------------------------------------------------------------------------
#| label: setup
#| include: false

required_pkgs <- c("glmmTMB", "lme4", "nlme")
have_all <- all(vapply(
  required_pkgs,
  requireNamespace,
  logical(1),
  quietly = TRUE
))
knitr::opts_chunk$set(eval = have_all)


## -----------------------------------------------------------------------------
#| label: simulated-data

library(seqwrap)
library(glmmTMB)


# Simulate data
dat <- seqwrap::simcounts(n_genes = 1000, 
                    n_samples = 50, 
                    beta_0 = 2, 
                    sigma_0 = 0.5, 
                    beta_1 = 0.5,, 
                    sigma_1 = 2, 
                    b_0 = 1,
                    clusters = 25,
                    overdispersion_min = c(1, 10))

# Calculate library sizes
dat$metadata$libsize <- log(colSums(dat$data[, -1]) )




## -----------------------------------------------------------------------------

swobject <- seqwrap_compose(
  data = dat$data, # A data frame with the first column as the target identifier
  metadata = dat$metadata, # Meta data bout samples, containing predictors
  samplename = "sample", # The name sample id column 
  modelfun = glmmTMB::glmmTMB, # The model fitting function
  # A list of argumnets supplied to the modelfunction
  arguments = list(
    formula = y ~ x + (1 | cluster),
    family = glmmTMB::nbinom1
  ),
  summary_fun = NULL,
  eval_fun = NULL,
  additional_vars = NULL
)



## -----------------------------------------------------------------------------

models <- seqwrap(swobject, 
                  return_models = TRUE,
                  subset = 1:10) 



## -----------------------------------------------------------------------------

modelsum <- seqwrap_summarise(models)



## -----------------------------------------------------------------------------
seqwrap::generic_summary


## -----------------------------------------------------------------------------
seqwrap::generic_evaluation


## -----------------------------------------------------------------------------

lib_sizes <- colSums(dat$data[,-1]) / 1e6

cpm <- sweep(dat$data[,-1], MARGIN = 2, 
             STATS = lib_sizes, 
             FUN = "/")

dat$data <- data.frame(cbind(dat$data[,1], cpm))



## -----------------------------------------------------------------------------

swobject <- seqwrap_compose(
  data = dat$data, # A data frame with the first column as the target identifier
  metadata = dat$metadata, # Meta data bout samples, containing predictors
  samplename = "sample", # The name sample id column 
  modelfun = lme4::lmer, # The model fitting function
  # A list of argumnets supplied to the modelfunction
  arguments = list(
    formula = y ~ x + (1 | cluster)
  ),
  summary_fun = NULL,
  eval_fun = NULL,
  additional_vars = NULL
)



## -----------------------------------------------------------------------------

models <- seqwrap(swobject, 
                  return_models = TRUE,
                  subset = 1:10) 



## -----------------------------------------------------------------------------

modelsum <- seqwrap_summarise(models)

modelsum$summaries



## -----------------------------------------------------------------------------

models_lme <- seqwrap(swobject, 
                  modelfun = nlme::lme,
                  
                  arguments = alist(fixed = y ~ x, 
                                    random = ~ 1|cluster), 
                  return_models = TRUE,
                  eval_fun = NULL,
                  subset = 1:10)

modelsum_lme <- seqwrap_summarise(models_lme)




## -----------------------------------------------------------------------------
lme_est <- data.frame(modelsum_lme$summaries[, c(1,4,5, 6,9)])
lmer_est <- data.frame(modelsum$summaries[, c(1,4,5, 6)])


## -----------------------------------------------------------------------------
#| echo: false
#| fig-cap: "Similar estimates between `lme4::lmer` and `nlme::lme` models"
#| label: fig-complme
plot(lme_est[lme_est$term == "x","estimate"], 
     lmer_est[lmer_est$term == "x","estimate"], 
     xlab = "nlme::lme estimate", 
     ylab = "lme4::lmer estimate")
abline(a = 0, b = 1)



## -----------------------------------------------------------------------------

models_gls <- seqwrap(swobject, 
                  modelfun = nlme::gls,
                  arguments = alist(
                  model = y ~ x,
                  correlation = nlme::corCompSymm(form = ~ 1|cluster) ),
                  additional_vars = "cluster",
                  return_models = TRUE,
                  eval_fun = NULL,
                  subset = 1:10)

modelsum_gls <- seqwrap_summarise(models_gls)



## -----------------------------------------------------------------------------


gls_est <- data.frame(modelsum_gls$summaries[, c(1,2,3,4,6)])




## -----------------------------------------------------------------------------
#| echo: false
#| fig-cap: "Similar estimates between `lme4::lmer` and `nlme::lme` models"
#| label: fig-compgls
plot(-log10(lme_est[lme_est$term == "x","p.value"]), 
     -log10(gls_est[gls_est$term == "x","p.value"]), 
     xlab = "nlme::lme p-value", 
     ylab = "nlme::gls p-value")
abline(a = 0, b = 1)

#' Generic summaries for model parameter estimates.
#'
#' When no summary function is provided to seqwrap (summary_fun), this function
#' uses broom(.mixed)::tidy to give a table of model parameter estimates.
#'
#' @param x A model fitted in seqwrap.
#'
#' @returns A tidy data frame possible to bind using seqwrap_summarise
#' @examples
#' # The generic summary works on model objects
#' library(seqwrap)
#'
#' if (requireNamespace("glmmTMB", quietly = TRUE)) {
#'   library(glmmTMB)
#'
#'   dat <- simcounts(n_genes = 1000,
#'                    n_samples = 30,
#'                    beta_0 = 5,
#'                    overdispersion_min_max = c(1, 10))
#'
#'   counts <- dat$data
#'   metadata <- dat$metadata
#'   metadata$ln_libsize <- log(colSums(counts[,-1]))
#'
#'   # Save counts for gene i in the same data frame
#'   metadata$y <- as.integer(counts[1, -1])
#'
#'   m <- glmmTMB(y ~ as.factor(x) + offset(ln_libsize),
#'                data = metadata,
#'                family = nbinom2)
#'
#'   generic_summary(m)
#'   generic_evaluation(m)
#' }
#'
#' @export
generic_summary <- function(x) {
  out <- broom.mixed::tidy(x)
  return(out)
}
#' Generic evaluation for model fits
#'
#' When no evaluation function is provided to seqwrap (eval_fun), this function
#' uses DHARMa to simulate residuals and test for uniformity, dispersion and
#' outliers. These tests could guide troubleshooting and detection of model
#' misspecification.
#'
#' @param x A model fitted in seqwrap.
#'
#' @returns A tidy data frame possible to combine using seqwrap_summarise
#' @examples
#' # The generic summary works on model objects
#' library(seqwrap)
#'
#' if (requireNamespace("glmmTMB", quietly = TRUE)) {
#'   library(glmmTMB)
#'
#'   dat <- simcounts(n_genes = 1000,
#'                    n_samples = 30,
#'                    beta_0 = 5,
#'                    overdispersion_min_max = c(1, 10))
#'
#'   counts <- dat$data
#'   metadata <- dat$metadata
#'   metadata$ln_libsize <- log(colSums(counts[,-1]))
#'
#'   # Save counts for gene i in the same data frame
#'   metadata$y <- as.integer(counts[1, -1])
#'
#'   m <- glmmTMB(y ~ as.factor(x) + offset(ln_libsize),
#'                data = metadata,
#'                family = nbinom2)
#'
#'   generic_summary(m)
#'   generic_evaluation(m)
#' }
#'
#' @export
generic_evaluation <- function(x) {
  # Simulate residuals using DHARMa
  sim_resid <- DHARMa::simulateResiduals(x, n = 250, plot = FALSE)
  # Combine potential metrics
  out <- tibble::tibble(
    uniformity = DHARMa::testUniformity(sim_resid, plot = FALSE)$p.value,
    dispersion = DHARMa::testDispersion(sim_resid, plot = FALSE)$p.value,
    outliers = DHARMa::testOutliers(sim_resid, plot = FALSE)$p.value
  )
  return(out)
}

# Defines custom property types for use in class constructors #

# Define a modelfun property that correctly updates model_print
modelfun_prop <- S7::new_property(
  validator = function(value) {
    if (!(is.null(value) || is.function(value))) "must be NULL or a function"
  }
)

# Accepts either NULL or a list, this will also validate data frames.
null_or_list <- S7::new_property(
  # The validator function checks if the value is NULL or a list
  validator = function(value) {
    if (!(is.null(value) || is.list(value))) "must be NULL or a list"
  }
)

# Define a custom property type that accepts either NULL or a list
null_or_df <- S7::new_property(
  # The validator function checks if the value is NULL or a list
  validator = function(value) {
    if (!(is.null(value) || is.data.frame(value)))
      "must be NULL or a data.frame"
  }
)


# Define a custom property type that accepts either data frame or a list
df_or_list <- S7::new_property(
  validator = function(value) {
    if (!(is.null(value) || is.data.frame(value) || is.list(value))) {
      "must be NULL, a data frame, or a list"
    }
  },
  default = NULL # Change default to NULL
)

# Empty data frame constant to avoid codoc mismatch from deparsing data.frame()
.empty_df <- data.frame()

# Define a call property for use in the swcontainer function
call_prop <- S7::new_property(
  validator = function(value) {
    if (!(is.call(value) || is.null(value))) {
      return("must be a language object representation of a call")
    }
    return(NULL) # Explicitly return NULL for valid values
  }
)

# Define a custom property type that accepts either NULL or function
null_or_function <- S7::new_property(
  validator = function(value) {
    if (!(is.null(value) || is.function(value))) "must be NULL or a function"
  }
)


#' seqwrapResults class
#'
#' seqwrapResults constructor function.
#'
#' @param .data A list containing initial data (inherited from S7::class_list)
#' @param models A list of fitted model objects
#' @param summaries A list of model summaries
#' @param evaluations A list of model evaluations
#' @param errors A data frame of errors and warnings. defaults to an empty dataframe
#' @param n Number of samples
#' @param k Number of targets
#' @param call_arguments Character string of function arguments used
#' @param call_engine Character string of modeling engine used
#' @param elapsed_time An proc.time() object giving the time needed to complete
#' iterative model fitting.
#'
#' @return An S7 object of class `seqwrap_results` storing fitted models,
#' summaries, evaluations, and diagnostic information from a `seqwrap()` run.
#'
#' @usage NULL
#'
#' @examples
#' # seqwrapResults is the S7 class returned by seqwrap(). End users do
#' # not normally call the constructor directly -- it is invoked
#' # internally once iterative model fitting has finished.
#'
#' library(seqwrap)
#'
#' dat <- simcounts(n_genes = 5)
#'
#' container <- seqwrap_compose(
#'   modelfun = stats::lm,
#'   arguments = list(formula = y ~ x),
#'   data = dat$data,
#'   metadata = dat$metadata,
#'   samplename = "sample"
#' )
#'
#' results <- seqwrap(container, subset = 1:2, cores = 1)
#'
#' # The object returned by seqwrap() is a seqwrapResults instance
#' S7::S7_inherits(results, seqwrapResults)
#'
#' @export
seqwrapResults <- S7::new_class(
  name = "seqwrap_results",
  parent = S7::class_list,
  properties = list(
    models = null_or_list,
    summaries = null_or_list,
    evaluations = null_or_list,
    errors = S7::new_property(S7::class_data.frame, default = .empty_df),
    n = S7::new_property(S7::class_numeric, default = integer(0)),
    k = S7::new_property(S7::class_numeric, default = integer(0)),
    call_arguments = S7::new_property(
      S7::class_character,
      default = character(0)
    ),
    call_engine = S7::new_property(S7::class_character, default = character(0)),
    elapsed_time = S7::new_property(S7::class_numeric, default = numeric(0))
  )
)


#' seqwrap container class
#'
#' The seqwrap container is used to store and validate data input
#' to the to seqwrap.
#'
#' @param .data A list containing initial data (inherited from S7::class_list)
#' @param modelfun A function used to fit models
#' @param arguments A list of arguments for the fitting function
#' @param data A data frame or list of data frames with target data
#' @param rownames Logical, should row names be used as target IDs?
#' @param metadata A data frame with sample information
#' @param targetdata A data frame with target-wise information
#' @param samplename Character for sample name identification
#' @param additional_vars Character vector of additional variables
#' @param summary_fun A function for summarizing models
#' @param eval_fun A function for evaluating models
#' @param exported A list of objects to export to workers
#' @param model_print Character representation of model function
#' @param arguments_print Character representation of arguments
#'
#' @return An S7 object of class `swcontainer` bundling data, metadata,
#' and the modelling function plus arguments to be consumed by `seqwrap()`.
#'
#' @usage NULL
#'
#' @examples
#' # swcontainer is the S7 class produced by seqwrap_compose(). End users
#' # normally build one via seqwrap_compose() rather than calling this
#' # constructor directly.
#'
#' library(seqwrap)
#'
#' dat <- simcounts(n_genes = 5)
#'
#' container <- seqwrap_compose(
#'   modelfun = stats::lm,
#'   arguments = list(formula = y ~ x),
#'   data = dat$data,
#'   metadata = dat$metadata,
#'   samplename = "sample"
#' )
#'
#' # The object returned by seqwrap_compose() is a swcontainer instance
#' S7::S7_inherits(container, swcontainer)
#'
#' @export
swcontainer <- S7::new_class(
  name = "swcontainer",
  parent = S7::class_list,
  properties = list(
    modelfun = modelfun_prop,
    arguments = null_or_list,
    data = null_or_list,
    rownames = S7::new_property(S7::class_logical, default = logical(0)),
    metadata = S7::new_property(S7::class_data.frame, default = .empty_df),
    targetdata = df_or_list,
    samplename = S7::new_property(S7::class_character, default = character(0)),
    additional_vars = S7::new_property(
      S7::class_character,
      default = character(0)
    ),
    summary_fun = null_or_function,
    eval_fun = null_or_function,
    exported = S7::new_property(S7::class_list, default = list()),
    model_print = S7::new_property(S7::class_character, default = character(0)),
    arguments_print = S7::new_property(
      S7::class_character,
      default = character(0)
    )
  )
)


#' Print method for objects of class seqwrapResults
#'
#' @description
#' Invoking the print method on seqwrapResults gives a summary
#' of the fitted objects.
#'
#' @param x A seqwrapResults object
#' @param ... Currently unused; included for compatibility with the
#'   `print` generic.
#'
#' @return Invisibly returns the object
#'
#' @usage NULL
#'
#' @examples
#' # Load packages and prepare data for examples --------------------------------
#'
#' library(seqwrap)
#'
#' if (requireNamespace("glmmTMB", quietly = TRUE)) {
#' library(glmmTMB)
#'
#' dat <- simcounts2()
#'
#' # Save simulated data as separate objects
#' counts <- dat$counts
#' metadata <- dat$metadata
#'
#' # Prepare library size for use as offset
#' metadata$ln_libsize <- log(metadata$library_size)
#'
#'
#' # A mixed effects negative binomial model of RNA-seq counts ------------------
#'
#' # Populate the seqwrap container
#' container <- seqwrap_compose(
#'   modelfun = glmmTMB::glmmTMB,
#'   arguments = list(
#'     formula = y ~ time * condition + (1|id) + offset(ln_libsize),
#'     family = glmmTMB::nbinom2()
#'   ),
#'   data = counts,
#'   metadata = metadata,
#'   samplename = "seq_sample_id"
#' )
#'
#' # Run seqwrap using the container
#' results <- seqwrap(container,
#'                    cores = 1)
#' print(results)
#' }
#'
#' @method print seqwrapResults
#' @name print.seqwrapResults
S7::method(print, seqwrapResults) <- function(x, ...) {
  cli::cli_h1("seqwrap")
  cli::cli_inform(
    "A total of {x@n} sample{?s} and {x@k} target{?s} where
                  used in {.code {x@call_engine}} with arguments
                  {.code {x@call_arguments}}"
  )

  # Count non-null values in the error/warning data frame
  errors_sum <- sapply(x@errors, function(x) sum(!sapply(x, is.null)))

  if (any(errors_sum[-1] > 0)) {
    cli::cli_alert_info("Some targets had associated errors or warnings")

    k <- x@k
    cli::cli_inform(c(
      "*" = "Fitting algorithm (errors): n = {errors_sum[2]}
      ({round(100 * (errors_sum[2]/k))}%)",
      "*" = "Fitting algorithm (warnings): n = {errors_sum[3]}
      ({round(100 * (errors_sum[3]/k))}%)",
      "*" = "Summary function (errors): n = {errors_sum[4]}
      ({round(100 * (errors_sum[4]/k))}%)",
      "*" = "Summary function (warnings): n = {errors_sum[5]}
      ({round(100 * (errors_sum[5]/k))}%)",
      "*" = "Evaluation function (errors): n = {errors_sum[6]}
      ({round(100 * (errors_sum[6]/k))}%)",
      "*" = "Evaluation function (warnings): n = {errors_sum[7]}
      ({round(100 * (errors_sum[7]/k))}%)"
    ))
  } else cli::cli_alert_info("No targets had associated errors or warnings")

  invisible(x)
}


#' Compose a swcontainer object for use in the seqwrap function.
#'
#' This function makes it possible to compose and  run checks on combined
#' data sets (meta data and target data) and fitting functions to avoid issues
#' in iterative modelling. See examples and vignettes for details.
#'
#' @param x An optional named list or DGEList object (DESeqDataSet not yet
#' implemented).
#' @param modelfun A model fitting function like stats::lm,
#' glmmTMB::glmmTMB or lme4::lmer
#' @param arguments An alist or list of arguments to be passed to the fitting
#'  function, this should not contain data. Note that the formula must have
#'  y as the dependent variable.
#' @param data A data frame or a list of data frames with targets (e.g. genes,
#' transcripts) as rows and sample names as columns.
#' If rownames = FALSE (default), each data frame should have target
#' identifications as the first column in the data frame(s). If rownames = TRUE
#' row names will be converted to target identifications. If data is provided as
#' a list, each element of the list should be named. The corresponding names
#' be available as variables for the fitting function.
#' @param rownames should row names in data be used as target identifications?
#' Defaults to FALSE.
#' @param metadata A data frame with sample names (corresponding to column
#' names in the target matrix)
#' and design variables.
#' @param targetdata A data frame or a list with target-wise values
#' (e.g. dispersion or start values) for each target. This data is made
#' available for the model fitting function and can be used to specify target
#' specific data in each iteration of seqwrap. When a data frame is provided
#' each row corresponds to the target specific value. When a list is provided,
#' each column of the data frame is available and can be called by name.
#' @param samplename A character value indicating the variable by which
#' metadata can merge with the target data. This defaults to "seq_sample_id"
#' as this is used in the trainomeMetaData package.
#' @param additional_vars A vector of additional variables that is contained
#' in the metadata data set that is needed to fit the model. By default the
#' metadata is reduced to variables contained in the slots
#' formula/model/fixed/random in additional arguments.
#' More variables may be needed for offsets, weights etc.
#' @param summary_fun A custom (user-created) function for
#' evaluating/summarizing models. If NULL, a list of fitted models are returned
#' @param eval_fun A custom (user-created) function for model
#' diagnostics/evaluation. If NULL, no evaluation/diagnostics of models are
#' returned
#' @param exported A list of functions, values etc. to be passed to
#' summary_fun and eval_fun. This list must contain any functions that
#' should be used in model summarise or evaluations.
#' @param update A list of named parameters to update a swcontainer object.
#' @return A swcontainer object for direct use in seqwrap.
#' @examples
#' # Load packages and prepare data for examples -------------------------------
#'
#' library(seqwrap)
#'
#' if (requireNamespace("glmmTMB", quietly = TRUE)) {
#' library(glmmTMB)
#'
#' dat <- simcounts2()
#'
#' # Save simulated data as separate objects
#' counts <- dat$counts
#' metadata <- dat$metadata
#'
#' # Prepare library size for use as offset
#' metadata$ln_libsize <- log(metadata$library_size)
#'
#'
#' # A mixed effects negative binomial model of RNA-seq counts -----------------
#'
#' # Populate the seqwrap container
#' container <- seqwrap_compose(
#'   modelfun = glmmTMB::glmmTMB,
#'   arguments = list(
#'     formula = y ~ time * condition + (1|id) + offset(ln_libsize),
#'     family = glmmTMB::nbinom2()
#'   ),
#'   data = counts,
#'   metadata = metadata,
#'   samplename = "seq_sample_id"
#' )
#'
#' # Run seqwrap using the container
#' results <- seqwrap(container,
#'                    cores = 1)
#'
#'
#'
#'
#' # Summarise results
#' summaries <- seqwrap_summarise(results)
#'
#'
#' # Including target-specific data --------------------------------------------
#'
#' # Target specific data can be supplied to seqwrap_compose to enable, e.g.,
#' # the use of priors for empirical Bayes shrinkage. In this example we are
#' # setting a dummy-prior on the `condition` parameter and target-specific
#' # prior on the dispersion parameter.
#'
#' fixef_priors <- data.frame(
#'   prior = "normal(0, 1)",
#'   class = "fixef",
#'   coef = "conditionB"
#' )
#'
#' dips_priors <- data.frame(
#'   prior = c(
#'     "normal(0.5, 0.25)",
#'     "normal(1, 0.25)"),
#'   class = "fixef_disp",
#'   coef = 1
#' )
#'
#'
#' # Combine information in target-specific list
#' prior_list <- list()
#' for(i in 1:2) {
#'
#'   prior_list[[i]] <- rbind(
#'     fixef_priors,
#'     dips_priors[i,]
#'   )
#'
#'
#' }
#'
#'
#'
#' container <- seqwrap_compose(
#'   modelfun = glmmTMB::glmmTMB,
#'   # NOTE: The use of `alist` prevents evaluation of list components
#'   arguments = alist(
#'     formula = y ~ time * condition + (1|id) + offset(ln_libsize),
#'     family = glmmTMB::nbinom2(),
#'     priors = data.frame(
#'       prior = prior,
#'       class = class,
#'       coef = coef
#'     )
#'   ),
#'   data = counts,
#'   metadata = metadata,
#'   targetdata = prior_list,
#'   samplename = "seq_sample_id"
#' )
#'
#' # Run seqwrap using the container
#' results_prior <- seqwrap(container,
#'                          # Return models to confirm use of prior information
#'                          # The use of prior information is not recommended
#'                          return_models = TRUE,
#'                          cores = 1)
#'
#' # Confirm prior information
#' summary(results_prior@models[[1]])
#' # Compare to naive model
#' summary(results@models[[1]])
#' }
#'
#' @export
seqwrap_compose <- function(
  x = NULL,
  modelfun,
  arguments,
  data,
  rownames = FALSE,
  metadata,
  targetdata = NULL,
  samplename = "seq_sample_id",
  additional_vars = NULL,
  summary_fun = NULL,
  eval_fun = NULL,
  exported = list(),
  update = list()
) {
  # Check if x is a swcontainer for updating
  if (S7::S7_inherits(x, swcontainer)) {
    # Pass updates as a list for the named parameter
    return(seqwrap_update(container = x, update))
  }

  # Extract the modelling algorithm for printing
  if (!is.null(modelfun)) {
    call_str <- match.call()
  }

  if (inherits(x, "DGEList")) {
    # Prepare count data from DGEList

    # If gene annotations are not available, use rownames
    if (is.null(x$genes)) {
      # If row names are numbers (1:nrow) add "target" for readability
      if (identical(rownames(x), as.character(1:nrow(x$counts)))) {
        xdata <- data.frame(
          target = paste0("target", 1:nrow(x$counts)),
          x$count
        )
        # Change name on sample names if provided
        if (!is.null(samplename)) colnames(xdata)[1] <- samplename
      } else {
        xdata <- data.frame(target = rownames(x$counts), x$count)
        # Change name on sample names if provided
        if (!is.null(samplename)) colnames(xdata)[1] <- samplename
      }
    }
    # If gene annotations are available, use these as identifiers
    if (!is.null(x$genes)) {
      xdata <- data.frame(target = x$genes[, 1], x$count)
      if (!is.null(samplename)) colnames(xdata)[1] <- samplename
    }

    # Prepare metadata from DGEList

    xmetadata <- data.frame(
      samplename = rownames(x$samples),
      x$samples,
      row.names = NULL
    )
    if (!is.null(samplename)) colnames(xmetadata)[1] <- samplename

    # Prepare gene-wise dispersion data

    container <- swcontainer(
      data = data_helper(data.frame(x$counts))
    )
  }

  # Initialize an empty container and add objects if they exist
  container <- swcontainer()

  # Set other properties
  if (!is.null(modelfun)) container@modelfun <- modelfun
  if (!is.null(modelfun)) container@model_print <- deparse(call_str$modelfun)
  if (!is.null(arguments)) container@arguments <- arguments
  if (!is.null(arguments))
    container@arguments_print <- deparse(call_str$arguments)
  if (!is.null(data)) container@data <- data
  if (!is.null(rownames)) container@rownames <- rownames
  if (!is.null(metadata)) container@metadata <- metadata
  if (!is.null(targetdata)) container@targetdata <- targetdata
  if (!is.null(samplename)) container@samplename <- samplename
  if (!is.null(additional_vars)) container@additional_vars <- additional_vars
  if (!is.null(summary_fun)) container@summary_fun <- summary_fun
  if (!is.null(eval_fun)) container@eval_fun <- eval_fun
  if (!is.null(exported)) container@exported <- exported

  # Validate targetdata
  if (!is.null(container@data) && !is.null(container@targetdata)) {
    n_data_rows <- nrow(container@data)
    n_target_rows <- if (is.data.frame(container@targetdata)) {
      nrow(container@targetdata)
    } else if (is.list(container@targetdata)) {
      length(container@targetdata)
    }

    if (n_data_rows != n_target_rows) {
      cli::cli_abort(
        "targetdata must have the same number of rows/elements ({n_target_rows})
      as data has rows ({n_data_rows})"
      )
    }
  }

  # Return the populated container
  return(container)
}


# Define a generic with named parameters only
seqwrap_update <- S7::new_generic("seqwrap_update", "container")

# Implement method for swcontainer class
S7::method(seqwrap_update, swcontainer) <- function(
  container,
  update = list()
) {
  if (!S7::S7_inherits(container, swcontainer)) {
    stop("First argument must be a swcontainer object")
  }

  if (length(update) == 0) {
    return(container) # Nothing to update
  }

  # Get property names
  name <- names(update)

  # Special handling for modelfun to update model_print
  if ("modelfun" %in% name) {
    # The model print information is not updated
    # due to difficulties getting the call from the
    # method. This is a temporary solution.
    update$model_print <- "The model function has been updated"
  }

  # Update the container object
  S7::props(container) <- update

  return(container)
}


#' Check swcontainer objects for seqwrap
#'
#' This function performs verbose checks/diagnostics of a swcontainer object.
#'
#' @param x A swcontainer object
#' @param verbose Logical, should the function print diagnostics? Default TRUE
#' @return `TRUE` (invisibly) when the container is consistent; otherwise
#' calls `cli::cli_abort()` with a descriptive message.
#' @keywords internal
#' @noRd
seqwrap_check <- function(x, verbose = TRUE) {
  # Check if the container is a swcontainer object
  if (!S7::S7_inherits(x, swcontainer)) {
    stop("The container must be a swcontainer object")
  }

  # checks if arguments for the provided fitting function matches arguments
  if (!all(names(x@arguments) %in% names(formals(x@modelfun)))) {
    cli::cli_abort(
      "Arguments do not match named arguments of the selected
                   \nmodel fitting function ('modelfun')."
    )
  }

  # Check if the data has a character vector for first column
  # (indicating transcript id).
  if (!is.character(x@data[, 1])) {
    cli::cli_abort(
      "The first column of the data is not character or
            factor,\ncheck if this column indicate target
            identifications."
    )
  }

  # Check if the sample name is present in the meta data
  if (!x@samplename %in% colnames(x@metadata)) {
    cli::cli_abort(
      "The samplename does not exist in the metadata,\nno
            variable for matching metadata and
            data."
    )
  }

  # Check that data is formatted correctly
  if (!all(x@metadata[, x@samplename] %in% names(x@data[, -1]))) {
    cli::cli_abort(
      "The sample names in the metadata does
         not match the sample column names in the seqdata."
    )
  }

  # Count the number of unique targets in the data
  if (x@rownames) {
    n_targets <- length(unique(rownames(x@data)))
  } else {
    n_targets <- length(unique(x@data[, 1]))
  }

  # If the target unique indicator is not the same as number of rows abort
  if (!n_targets == nrow(x@data)) {
    cli::cli_abort(
      "The number of unique target identifications does not
        match the number of rows in the data. You might have duplicate id's."
    )
  }

  # Count the number of samples
  n_samples <- length(unique((x@metadata[, x@samplename])))

  # If the unique sample indicator is not the same as number of rows abort
  if (!n_samples == nrow(x@metadata)) {
    cli::cli_abort(
      "The number of unique sample identifications does not
        match the number of rows in the meta data.
      You might have duplicate sample id's."
    )
  }

  # Check that sample id match between data and meta data
  meta_data_sample_id <- unique((x@metadata[, x@samplename]))

  if (x@rownames) {
    data_sample_id <- colnames(x@data)
  } else {
    data_sample_id <- colnames(x@data[, -1])
  }

  # Save sample id for printing
  if (length(data_sample_id) > 3) {
    sample_id_print <- paste(data_sample_id[1:3], collapse = ", ")
  } else {
    sample_id_print <- paste0(paste(data_sample_id, collapse = ", "), ", ...")
  }

  if (
    !all(meta_data_sample_id %in% data_sample_id) &&
      all(data_sample_id %in% meta_data_sample_id)
  ) {
    cli::cli_abort(
      "The sample names in the meta data does not match the sample column names
    in the data. Check if the sample names are correctly formatted and that
    `samplename` looks for the correct variable in the meta data."
    )
  }

  # If verbose is TRUE, print the diagnostics
  if (verbose) {
    printfun <- function() {
      cli::cli_h1("seqwrap diagnostics")

      cli::cli_inform(
        "The swcontainer object has been checked for
                      consistency with the following diagnostics:\n"
      )

      cli::cli_ul(
        "The number of unique target identifications is {n_targets}."
      )

      cli::cli_ul(
        "The number of unique sample identifications is {n_samples}."
      )

      cli::cli_ul(
        "Sample identifications ({sample_id_print}) match between data
        and meta data."
      )

      cli::cli_ul(
        "The modelling function to be used is {x@model_print}."
      )
    }
    printfun()
  }

  # If the swcontainer is a fully populated data container return TRUE
  return(TRUE)
}


# #' Print method for objects of class swcontainer
# #'
# #' @description
# #' Invoking the print method on swcontainer gives a summary
# #' of data container.
# #'
# #' @param x A swcontainer object
# #'
# #' @return Invisibly returns the object
# #'
# #' @examples
# #' seqwrap.dat <- seqwrap_compose(...)
# #' print(seqwrap.dat)
# #'
# #' @method print swcontainer
# #' @name print.swcontainer
# S7::method(print, swcontainer) <- function(x) {
#   cli::cli_h1("seqwrap data container")
#
#   if (!is.null(x@fitting_fun)) {
#
#   }
#
#
#  # invisible(x)
# }

#' A flexible upper-level wrapper for iterative modelling using any available
#' fitting algorithm
#'
#'
#' @inheritParams seqwrap_compose
#' @param y, An swcontainer object created with seqwrap_compose, a named list
#' or a DGEList object.
#' @param return_models Logical, should models be returned as part of the
#' output? Save models during development on subsets of the data.
#' If used on large data sets, this will result in large memory usage.
#' @param save_models Logical, should models be saved? Models may be saved
#' on disk to save working memory.
#' @param model_path A character. The path to saved models.
#' @param subset A sequence, random samples or integers to indicate which
#' rows to keep in data. This is useful if you want to test the model in a
#' subset of targets. If left to the default (NULL), all rows will be used.
#' @param cores An integer indicating the number of cores to be used in parallel
#'  computations. If NULL, a sequential for loop is used. If "max", all
#'  available cores are used.
#' @param verbose Logical, should the function print diagnostics after checking
#' the data container?
#' @return A nested list with three upper levels slots: models, a list of
#' fitted objects; summaries, a list of summaries created from the summary_fun
#' function; evaluations, a list of diagnostics created from eval_fun.
#' @details This function provides a flexible wrapper to fit, summarize and
#' evaluate statistical models fitted to high dimensional omics-type data.
#' Models are fitted and passed to user defined functions to summarize and
#' evaluate models.
#' @examples
#' library(seqwrap)
#'
#' if (requireNamespace("glmmTMB", quietly = TRUE)) {
#' library(glmmTMB)
#'
#' # Simulate n targets
#' dat <- simcounts(n_genes = 10000)
#'
#' # Save simulated data as separate objects
#' counts <- dat$data
#' metadata <- dat$metadata
#'
#' # Prepare library size for use as offset
#' metadata$ln_libsize <- log(colSums(counts[,-1]))
#'
#'
#' # A mixed effects negative binomial model of RNA-seq counts -----------------
#'
#' # Populate the seqwrap container
#' container <- seqwrap_compose(
#'   modelfun = glmmTMB::glmmTMB,
#'   arguments = list(
#'     formula = y ~ x + (1|cluster) + offset(ln_libsize),
#'     family = glmmTMB::nbinom2()
#'   ),
#'   data = counts,
#'   metadata = metadata,
#'   samplename = "sample"
#' )
#'
#' # Run seqwrap using the container on a subset of targets
#' results <- seqwrap(container,
#'                    subset = 1:5,
#'                    cores = 1)
#'
#'
#' # Summarise results only contains the subset
#' summaries <- seqwrap_summarise(results)
#' }
#'
#' @export
seqwrap <- function(
  y = NULL,
  modelfun = NULL,
  arguments = NULL,
  data = NULL,
  metadata = NULL,
  samplename = NULL,
  additional_vars = NULL,
  summary_fun = NULL,
  eval_fun = NULL,
  exported = list(),
  return_models = TRUE,
  save_models = FALSE,
  model_path = NULL,
  subset = NULL,
  cores = 1,
  verbose = TRUE
) {
  # Elapsed time
  start_time <- proc.time()

  # If the input is a swcontainer object, use the object
  if (S7::S7_inherits(y, seqwrap::swcontainer)) {
    container <- y

    # If variables are to be updated, update the container
    updates <- list()

    if (!is.null(modelfun)) updates$modelfun <- modelfun
    if (!is.null(arguments)) updates$arguments <- arguments
    if (!is.null(data)) updates$data <- data
    if (!is.null(metadata)) updates$metadata <- metadata
    if (!is.null(samplename)) updates$samplename <- samplename
    if (!is.null(additional_vars)) updates$additional_vars <- additional_vars
    if (!is.null(summary_fun)) updates$summary_fun <- summary_fun
    if (!is.null(eval_fun)) updates$eval_fun <- eval_fun
    if (!is.null(exported)) updates$exported <- exported

    # Update the container
    if (length(updates) > 0)
      container <- seqwrap_compose(container, update = updates)
  } else if (inherits(y, "DGEList")) {
    # If the input is a DGEList object, compose a new container
    container <- seqwrap_compose(x = y)

    # If variables are to be updated, update the container
    updates <- list()

    if (!is.null(modelfun)) updates$modelfun <- modelfun
    if (!is.null(arguments)) updates$arguments <- arguments
    if (!is.null(data)) updates$data <- data
    if (!is.null(metadata)) updates$metadata <- metadata
    if (!is.null(samplename)) updates$samplename <- samplename
    if (!is.null(additional_vars)) updates$additional_vars <- additional_vars
    if (!is.null(summary_fun)) updates$summary_fun <- summary_fun
    if (!is.null(eval_fun)) updates$eval_fun <- eval_fun
    if (!is.null(exported)) updates$exported <- exported

    # Update the container
    container <- seqwrap_compose(container, update = updates)
  } else if (is.null(y)) {
    # If the input is NULL, compose a new container
    container <- seqwrap_compose(
      modelfun = modelfun,
      arguments = arguments,
      data = data,
      metadata = metadata,
      samplename = samplename,
      additional_vars = additional_vars,
      summary_fun = summary_fun,
      eval_fun = eval_fun,
      exported = exported
    )
  } else {
    stop("The input must be a swcontainer object, a DGEList object or NULL")
  }

  # If eval_fun or summary_fun are NULL, supply generic functions
  if (is.null(container@summary_fun)) container@summary_fun <- generic_summary
  if (is.null(container@eval_fun)) container@eval_fun <- generic_evaluation

  # If subset is provided, subset the data
  if (!is.null(subset)) {
    # Subset the data
    container@data <- container@data[subset, , drop = FALSE]

    # Subset the targetdata if it is not NULL
    if (!is.null(container@targetdata)) {
      if (is.data.frame(container@targetdata)) {
        container@targetdata <- container@targetdata[subset, , drop = FALSE]
      } else if (is.list(container@targetdata)) {
        container@targetdata <- container@targetdata[subset]
      }
    }
  }

  # Validate model_path when save_models is TRUE
  if (save_models && is.null(model_path)) {
    stop("'model_path' must be specified when 'save_models' is TRUE")
  }

  # Check the container for consistency
  # seqwrap_check(container, verbose = verbose)

  # Get the number of targets and samples
  k <- nrow(container@data)
  n <- nrow(container@metadata)

  # data_helper function. Combine data into a list of data frames
  # containing variables, y in case of user-provided data frame;
  # names of variables from list names in case of user-provided
  # named list.

  if (is.null(container@targetdata)) {
    dfs <- data_helper(container@data)
  } else {
    dfs <- data_helper(container@data, container@targetdata)
  }

  # Determine the number of cores
  if (is.null(cores)) {
    num_cores <- 1L
  } else if (is.character(cores) && identical(cores, "max")) {
    num_cores <- parallel::detectCores()
  } else if (is.numeric(cores) && length(cores) == 1L) {
    num_cores <- min(as.integer(cores), parallel::detectCores())
  } else {
    stop("'cores' must be NULL, a single integer, or the string \"max\"")
  }

  # Catch the function calls for printing
  funcall <- match.call()
  ## Arguments to string
  call_arguments <- deparse(funcall$arguments)
  call_arguments <- sub("^.*?\\((.*)\\).*$", "\\1", call_arguments)
  ## modelfun to string
  call_engine <- deparse(funcall$modelfun)

  # Print pre-fit information

  if (verbose) {
    cli::cli_h1("seqwrap")
    cli::cli_inform(
      "Initiating clusters for parallel processing with {num_cores} core{?s}"
    )
  }

  ## Applying the model function in parallel ##

  # Making variables from the container available in the parallel environment
  metadata <- container@metadata
  arguments <- container@arguments
  modelfun <- container@modelfun
  samplename <- container@samplename
  additional_vars <- container@additional_vars
  summary_fun <- container@summary_fun
  eval_fun <- container@eval_fun
  exported <- container@exported

  # Create a cluster using the number of cores specified
  cl <- parallel::makeCluster(num_cores)

  # Load the seqwrap namespace (and its Imports) on each worker
  parallel::clusterEvalQ(cl, requireNamespace("seqwrap"))

  ## Export data to clusters
  parallel::clusterExport(
    cl,
    varlist = c(
      "metadata",
      "arguments",
      "modelfun",
      "samplename",
      "additional_vars",
      "save_models",
      "exported",
      "model_path",
      "return_models",
      "summary_fun",
      "eval_fun"
    ),
    envir = environment()
  )

  # Parallel execution of the fitting process
  if (verbose) {
    cli::cli_inform("Merging and modelling data")
  }

  results <- pbapply::pblapply(
    cl = cl,
    X = dfs,
    FUN = seqwrap_mtf,
    samp_name = samplename,
    metdat = metadata,
    arg_list = arguments,
    add_vars = additional_vars,
    mt_summary_fun = summary_fun,
    mt_eval_fun = eval_fun,
    ffun = modelfun,
    return_mod = return_models,
    save_mods = save_models,
    mod_path = model_path
  )

  parallel::stopCluster(cl)

  # Combine results

  models <- NULL
  summaries <- NULL
  evaluations <- NULL
  errors <- NULL

  # Collect models
  if (return_models) models <- lapply(results, `[[`, "model")

  if (!is.null(summary_fun)) summaries <- lapply(results, `[[`, "summaries")
  if (!is.null(eval_fun)) evaluations <- lapply(results, `[[`, "evaluation")

  ## Create a data frame of all errors/warnings
  errors <- tibble::tibble(
    target = names(results),
    errors_fit = lapply(results, `[[`, "err"),
    warnings_fit = lapply(results, `[[`, "warn"),
    err_sum = lapply(results, `[[`, "err_sum"),
    warn_sum = lapply(results, `[[`, "warn_sum"),
    err_eval = lapply(results, `[[`, "err_eval"),
    warn_eval = lapply(results, `[[`, "warn_eval")
  )

  # Count non-null values in the error/warning data frame
  errors_sum <- sapply(errors, function(x) sum(!sapply(x, is.null)))

  # Track the elapsed time
  elapsed_time <- proc.time() - start_time
  sys_time <- Sys.time()

  ## Evaluate errors for the resulting print function
  if (verbose) {
    cli::cli_inform({
      paste0(
        format(sys_time, "%b %d %Y %X"),
        ": Completed model fitting and evaluation. Elapsed time was ",
        round(elapsed_time[[3]] / 60, 2),
        " minutes"
      )
    })

    if (any(errors_sum[-1] > 0)) {
      cli::cli_alert_info("Some targets had associated errors or warnings")

      cli::cli_inform(c(
        "*" = "Modeling algorithm (errors): n = {errors_sum[2]}
      ({round(100 * (errors_sum[2]/k))}%)",
        "*" = "Modeling algorithm (warnings): n = {errors_sum[3]}
      ({round(100 * (errors_sum[3]/k))}%)",
        "*" = "Summary function (errors): n = {errors_sum[4]}
      ({round(100 * (errors_sum[4]/k))}%)",
        "*" = "Summary function (warnings): n = {errors_sum[5]}
      ({round(100 * (errors_sum[5]/k))}%)",
        "*" = "Evaluation function (errors): n = {errors_sum[6]}
      ({round(100 * (errors_sum[6]/k))}%)",
        "*" = "Evaluation function (warnings): n = {errors_sum[7]}
      ({round(100 * (errors_sum[7]/k))}%)"
      ))
    }
  }

  ## Combine the results into a seqwrapResults
  comb_results <- seqwrapResults(
    models = models,
    summaries = summaries,
    evaluations = evaluations,
    errors = errors,
    n = n,
    k = k,
    call_arguments = container@arguments_print,
    call_engine = container@model_print,
    elapsed_time = elapsed_time
  )

  return(comb_results)
}


#' Summarise seqwrapResults objects
#'
#' @param x A seqwrapResults object
#' @param summaries Logical, should summaries be combined?
#' @param evaluations Logical, should evaluations be combined?
#' @param verbose, Logical should progress be printed? Default TRUE
#' @param errors Logical, should errors be combined?
#'
#' @details
#' This functions attempts to summarize results from the summary and evaluation
#' functions applied in each iteration during modelling. The function expects
#' that the summary and evaluation functions return data frames.
#'
#' @return A list (invisibly) with up to two data frames: `summaries`,
#' combined parameter summaries from each model, and `evaluations`, combined
#' diagnostics from each model. Entries are omitted when the corresponding
#' slot of `x` is empty or the user disables them via the `summaries` /
#' `evaluations` arguments.
#' @examples
#' # Load packages and prepare data for examples -------------------------------
#'
#' library(seqwrap)
#'
#' if (requireNamespace("glmmTMB", quietly = TRUE)) {
#' library(glmmTMB)
#'
#' # Simulate n targets
#' dat <- simcounts(n_genes = 10000)
#'
#' # Save simulated data as separate objects
#' counts <- dat$data
#' metadata <- dat$metadata
#'
#' # Prepare library size for use as offset
#' metadata$ln_libsize <- log(colSums(counts[,-1]))
#'
#'
#' # A mixed effects negative binomial model of RNA-seq counts -----------------
#'
#' # Populate the seqwrap container
#' container <- seqwrap_compose(
#'   modelfun = glmmTMB::glmmTMB,
#'   arguments = list(
#'     formula = y ~ x + (1|cluster) + offset(ln_libsize),
#'     family = glmmTMB::nbinom2()
#'   ),
#'   data = counts,
#'   metadata = metadata,
#'   samplename = "sample"
#' )
#'
#' # Run seqwrap using the container on a subset of targets
#' results <- seqwrap(container,
#'                    subset = 1:5,
#'                    cores = 1)
#'
#'
#' # Summarise results only contains the subset
#' summaries <- seqwrap_summarise(results)
#'
#' # Get summaries
#' summaries$summaries
#'
#' # Get model evaluations
#' summaries$evaluations
#' }
#'
#'
#' @export
seqwrap_summarise <- function(
  x,
  summaries = TRUE,
  evaluations = TRUE,
  errors = TRUE,
  verbose = TRUE
) {
  # Check if the input is a seqwrapResults object
  if (!S7::S7_inherits(x, seqwrapResults)) {
    stop("The input must be a seqwrapResults object")
  }

  ## Print information from the seqwrapResults object
  print_info <- function() {
    cli::cli_h1("seqwrap summarise")

    cli::cli_li(
      "A total of {x@n} sample{?s} and {x@k} target{?s} where
                  used in {.code {x@call_engine}} with arguments
                  {.code {x@call_arguments}}"
    )
    if (summaries) {
      cli::cli_li(
        "Attempting to combine results from
                  the provided summary function."
      )
    }

    if (evaluations) {
      cli::cli_li(
        "Attempting to combine results from
                  the provided evaluations function."
      )
    }

    if (errors) {
      cli::cli_li(
        "Attempting to summarise errors and
                  warnings from the fitting process."
      )
    }
  }

  if (verbose) print_info()

  ## Initialize results variables ##
  summarised_results_final <- NULL
  evaluated_results_final <- NULL

  # print(x)

  ## Extract summarises
  if (summaries) {
    # Check for NULL or empty list
    if (is.null(x@summaries) || all(sapply(x@summaries, is.null))) {
      print_summary <- function() {
        cli::cli_h1("Model summaries")
        cli::cli_alert_info("No summaries available")
      }
      if (verbose) print_summary()
    } else {
      # Count the number of non-null elements in the list of summaries
      n_summaries <- sum(!sapply(x@summaries, is.null))

      print_summary <- function() {
        cli::cli_h1("Model summaries")
        cli::cli_li("{n_summaries} targets have associated summaries")
      }

      if (verbose) print_summary()

      # Filter out NULL elements from the list of summaries
      x@summaries <- x@summaries[!sapply(x@summaries, is.null)]

      # Use Map to put names in the target column
      temp_list_summaries <- Map(
        function(df, df_name) {
          df[["target"]] <- df_name
          # Move the target column to the first column
          df <- df[, c("target", setdiff(colnames(df), "target")), drop = FALSE]
          # Return data frames
          df
        },
        x@summaries,
        names(x@summaries)
      )

      # Bind the list of data frames
      summarised_results_final <- do.call(rbind, temp_list_summaries)
      # NOW set rownames to NULL on the combined data frame
      rownames(summarised_results_final) <- NULL
    }
  } # End if (summaries)

  ## Extract evaluations
  if (evaluations) {
    # Check for NULL or empty list
    if (is.null(x@evaluations) || all(sapply(x@evaluations, is.null))) {
      print_evaluation <- function() {
        cli::cli_h1("Model evaluations")
        cli::cli_alert_info("No evaluations available")
      }
      if (verbose) print_evaluation()
      # evaluated_results_final remains NULL
    } else {
      # Count the number of non-null elements in the list of evaluations
      n_evals <- sum(!sapply(x@evaluations, is.null))

      print_evaluation <- function() {
        cli::cli_h1("Model evaluations")
        cli::cli_li("{n_evals} targets have associated evaluation results")
      }

      if (verbose) print_evaluation()

      # Filter out NULL elements from the list of evaluations
      x@evaluations <- x@evaluations[!sapply(x@evaluations, is.null)]

      # Use Map to process each data frame in the list
      temp_list_evaluations <- Map(
        function(df, df_name) {
          df[["target"]] <- df_name
          # Move the target column to the first column
          df <- df[, c("target", setdiff(colnames(df), "target")), drop = FALSE]

          df
        },
        x@evaluations,
        names(x@evaluations)
      )

      # Bind the list of data frames
      evaluated_results_final <- do.call(rbind, temp_list_evaluations)
      rownames(evaluated_results_final) <- NULL
    }
  } # End if (evaluations)

  ## Combine results
  # Create empty results list
  results <- list()

  if (!is.null(summarised_results_final)) {
    results$summaries <- summarised_results_final
  }

  if (!is.null(evaluated_results_final)) {
    results$evaluations <- evaluated_results_final
  }

  # Add error handling results here if implemented

  # Check if the final list is empty
  if (length(results) == 0) {
    cli::cli_alert_warning("No results were generated, check your input.")
  }

  # Check if the final list is empty
  if (length(results) != 0) {
    if (verbose) {
      cli::cli_h1("Combined results")
      cli::cli_alert_info(
        "Combined results have been generated and were
                        silently returned."
      )
    }
  }

  # Return the final list object
  return(invisible(results))
}

# Packages
library(testthat)
library(seqwrap)


test_that("seqwrap returns a list of models in the model
          slot when asked to return models", {
  # Simulate data
  dat <- seqwrap::simcounts(
    seed = 1,
    n_samples = 40,
    n_genes = 10,
    clusters = 20
  )

  swobject <- seqwrap::seqwrap_compose(
    modelfun = glmmTMB::glmmTMB,
    arguments = list(
      formula = y ~
        x +
          (1 | cluster),
      family = glmmTMB::nbinom1
    ),
    data = dat$data,
    metadata = dat$metadata,
    samplename = "sample",
    additional_vars = NULL
  )

  test_glmmtmb <- seqwrap::seqwrap(swobject, return_models = TRUE, cores = 2)

  expect_s3_class(test_glmmtmb@models[[1]], "glmmTMB")

  test_glmnb <- seqwrap::seqwrap(
    swobject,
    arguments = list(formula = y ~ x),
    modelfun = MASS::glm.nb,
    return_models = TRUE,
    cores = 1
  )

  expect_s3_class(test_glmnb@models[[1]], c("glm", "lm", "negbin"))

  test_lm <- seqwrap::seqwrap(
    swobject,
    modelfun = stats::lm,
    arguments = list(formula = y ~ x),
    return_models = TRUE,
    cores = 1
  )

  expect_s3_class(test_lm@models[[1]], "lm")

  test_glm <- seqwrap::seqwrap(
    swobject,
    modelfun = stats::glm,
    arguments = alist(
      formula = y ~ x,
      family = poisson(link = "log")
    ),
    return_models = TRUE,
    cores = 1
  )

  expect_s3_class(test_glm@models[[1]], "glm")

  test_lme <- seqwrap::seqwrap(
    swobject,
    modelfun = nlme::lme,
    arguments = alist(
      fixed = y ~ x,
      random = ~ 1 | cluster
    ),
    return_models = TRUE,
    cores = 1
  )

  expect_s3_class(test_lme@models[[1]], "lme")
})


test_that("Model summaries and evaluations returns expected results", {
  ## Create a model summary function
  summaryfun_glmmtmb <- function(x) {
    # Extract conditional effects and store as a tibble
    results <- tibble::as_tibble(coef(summary(x))$cond, rownames = "coef") |>
      dplyr::select(
        coef,
        estimate = Estimate,
        se = "Std. Error",
        z = "z value",
        p = "Pr(>|z|)"
      )

    ## Return results
    return(results)
  }

  evalfun_glmmtmb <- function(x) {
    simresid <- DHARMa::simulateResiduals(
      fittedModel = x,
      plot = FALSE,
      n = 1000
    )

    unif <- DHARMa::testUniformity(simresid, plot = FALSE)
    disp <- DHARMa::testDispersion(simresid, plot = FALSE)
    pdhess <- x$sdr$pdHess

    return(data.frame(
      unif = unif$p.value,
      disp = disp$p.value,
      pdhess = pdhess
    ))
  }

  # Simulate data
  dat <- seqwrap::simcounts(
    seed = 1,
    n_samples = 40,
    n_genes = 10,
    clusters = 20
  )

  swobject <- seqwrap::seqwrap_compose(
    modelfun = glmmTMB::glmmTMB,
    arguments = list(
      formula = y ~
        x +
          (1 | cluster),
      family = glmmTMB::nbinom1
    ),
    data = dat$data,
    metadata = dat$metadata,
    samplename = "sample",
    additional_vars = NULL
  )

  testsummary_glmmtmb <- seqwrap::seqwrap(
    swobject,
    summary_fun = summaryfun_glmmtmb,
    eval_fun = evalfun_glmmtmb,
    cores = 1
  )

  ## Expected output from summary/evaluation functions
  expect_s3_class(testsummary_glmmtmb@summaries[[1]], "tbl")
  expect_s3_class(testsummary_glmmtmb@evaluations[[1]], "data.frame")

  ## Expect no errors in the error data
  expect_null(testsummary_glmmtmb@errors$err_sum[[1]])
  expect_null(testsummary_glmmtmb@errors$err_eval[[1]])

  # TODO add tests for the other model types
})


test_that("seqwrap is silent when setting verbose = FALSE", {
  # Simulate data
  dat <- seqwrap::simcounts(
    seed = 1,
    n_samples = 40,
    n_genes = 10,
    clusters = 20
  )

  swobject <- seqwrap::seqwrap_compose(
    modelfun = glmmTMB::glmmTMB,
    arguments = list(
      formula = y ~
        x +
          (1 | cluster),
      family = glmmTMB::nbinom1
    ),
    data = dat$data,
    metadata = dat$metadata,
    samplename = "sample",
    additional_vars = NULL
  )

  expect_no_message({
    test_glmmtmb <- seqwrap::seqwrap(
      swobject,
      return_models = TRUE,
      verbose = FALSE,
      cores = 1
    )
  })

  expect_no_message({
    temp <- seqwrap_summarise(test_glmmtmb, verbose = FALSE)
  })
})


test_that("seqwrap can handle complex random effects", {
  # Simulate data
  dat1 <- seqwrap::simcounts(
    seed = 1,
    n_samples = 160,
    n_genes = 10,
    clusters = 80
  )
  dat2 <- seqwrap::simcounts(
    seed = 1,
    n_samples = 40,
    n_genes = 10,
    clusters = 20
  )

  metdata1 <- dat1$metadata |>
    dplyr::filter(cluster %in% paste0("c", 60:100)) |>
    dplyr::mutate(study = "s1", cluster = paste0("A", cluster))

  count1 <- dat1$data |>
    dplyr::select(targetid, tidyselect::matches(metdata1$sample))

  data <- cbind(count1, dat2$data[, -1])
  metadata <- metdata1 |>
    dplyr::bind_rows(
      dat2$metadata |>
        dplyr::mutate(study = "s2")
    )

  swobject <- seqwrap::seqwrap_compose(
    modelfun = glmmTMB::glmmTMB,
    arguments = alist(
      formula = y ~ x + (1 | cluster) + (1 + x || study),
      family = glmmTMB::nbinom1
    ),
    data = data,
    metadata = metadata,
    samplename = "sample",
    additional_vars = NULL
  )

  expect_no_message({
    test_glmmtmb <- seqwrap::seqwrap(
      swobject,
      return_models = TRUE,
      verbose = FALSE,
      cores = 1
    )
  })

  ## Check that all models are fitted
  expect_equal(length(test_glmmtmb@models), 10)
})

# Use this for specific tests of lme functionality
#
#
#
library(testthat)
library(seqwrap)

test_that("seqwrap lme and gls models from the nlme package", {

            # Simulate data
            dat <- seqwrap::simcounts(seed = 1,
                                       n_samples = 40,
                                       n_genes = 10,
                                       clusters = 20)

            swobject <- seqwrap::seqwrap_compose(
              modelfun = nlme::lme,
              arguments = alist(
                fixed = y ~ x,
                random = ~ 1|cluster
              ),
              data = dat$data,
              metadata = dat$metadata,
              samplename = "sample",
              additional_vars = NULL
            )


            test_lme <- seqwrap::seqwrap(
              swobject,
              return_models = TRUE,
              cores = 1
            )

            expect_s3_class(test_lme@models[[1]], "lme")



            test_gls <- seqwrap::seqwrap(
              swobject,
              modelfun = nlme::gls,
              arguments = alist(
                model = y ~ x,
                correlation = nlme::corCompSymm(form = ~ 1|cluster) ),
              additional_vars = "cluster",
              return_models = TRUE,
              cores = 1
            )


            expect_s3_class(test_gls@models[[1]], "gls")




          })

library(testthat)
library(seqwrap)


test_that("seqwrap accepts NULL (the default) as target data", {

  # Simulate data
  dat <- seqwrap::simcounts(seed = 1,
                            n_samples = 40,
                            n_genes = 10,
                            clusters = 20)


 expect_no_error(
   seqwrap::seqwrap_compose(
   modelfun = glmmTMB::glmmTMB,
   # No targetdata -- Explicit NULL
   targetdata = NULL,
   arguments = alist(
     formula = y ~
       x +
       (1 | cluster),
     family = glmmTMB::nbinom2,
     priors = data.frame(prior = Prior,
                         class = Class,
                         coef = Coef)
   ),
   data = dat$data,
   metadata = dat$metadata,
   samplename = "sample",
   additional_vars = NULL
 ))


 expect_no_error(
   seqwrap::seqwrap_compose(
     modelfun = glmmTMB::glmmTMB,
  # No target data -- Implicit null
  # targetdata = NULL,
     arguments = alist(
       formula = y ~
         x +
         (1 | cluster),
       family = glmmTMB::nbinom2,
       priors = data.frame(prior = Prior,
                           class = Class,
                           coef = Coef)
     ),
     data = dat$data,
     metadata = dat$metadata,
     samplename = "sample",
     additional_vars = NULL
   ))



})








test_that("seqwrap returns a list of models when target data (data frame) is used to pass
          variables used in arguments.", {

            # Simulate data
            dat <- seqwrap::simcounts(seed = 1,
                                       n_samples = 40,
                                       n_genes = 10,
                                       clusters = 20)

            # Target data created with variables
            tardat <- data.frame(Prior = paste0("normal(",
                                                rnorm(10, 2, 0.1),
                                                ",",
                                                rnorm(10, 0.3, 0.001),
                                                ")"),
                                 Class = rep("fixef_disp", 10),
                                 Coef = rep(1, 10))


            swobject <- seqwrap::seqwrap_compose(
              modelfun = glmmTMB::glmmTMB,
              # Target data with multiple variables
              targetdata = tardat,
              arguments = alist(
                formula = y ~
                  x +
                  (1 | cluster),
                family = glmmTMB::nbinom2,
                priors = data.frame(prior = Prior,
                                   class = Class,
                                   coef = Coef)
              ),
              data = dat$data,
              metadata = dat$metadata,
              samplename = "sample",
              additional_vars = NULL
            )





            test_glmmtmb <- seqwrap::seqwrap(swobject,
                                             return_models = TRUE,
                                             cores = 1)

            # Subset test
            test_glmmtmb_subset <- seqwrap::seqwrap(swobject,
                                             return_models = TRUE,
                                             subset = 1:5,
                                             cores = 1)

            expect_s3_class(test_glmmtmb@models[[1]], "glmmTMB")
            expect_s3_class(test_glmmtmb_subset@models[[1]], "glmmTMB")

            # Check that glmmTMB has used priors
            expect_no_error(test_glmmtmb@models[[1]]$modelInfo$priors)
            expect_no_error(test_glmmtmb_subset@models[[1]]$modelInfo$priors)



          })


test_that("seqwrap returns a list of models when target data (list) is used
to pass variables used in arguments.", {

            # Simulate data
            dat <- seqwrap::simcounts(seed = 1,
                                      n_samples = 40,
                                      n_genes = 10,
                                      clusters = 20)

            # Target data created with variables
            tardat <- data.frame(prior = c("normal(0, 1)","normal(2, 1)"),
                                 class = c("beta", "fixef_disp"),
                                 coef = c("x", "1"))

            tardat_list <- list()

            for(i in 1:10) {
              tardat_list[[i]] <- tardat
            }




            swobject <- seqwrap::seqwrap_compose(
              modelfun = glmmTMB::glmmTMB,
              # Target data with multiple variables
              targetdata = tardat_list,
              arguments = alist(
                formula = y ~
                  x +
                  (1 | cluster),
                family = glmmTMB::nbinom2,
                priors = data.frame(prior = prior,
                                    class = class,
                                    coef = coef)
              ),
              data = dat$data,
              metadata = dat$metadata,
              samplename = "sample",
              additional_vars = NULL
            )





            test_glmmtmb <- seqwrap::seqwrap(swobject,
                                             return_models = TRUE,
                                             cores = 1)

            # Subset test
            test_glmmtmb_subset <- seqwrap::seqwrap(swobject,
                                                    return_models = TRUE,
                                                    subset = 1:5,
                                                    cores = 1)

            expect_s3_class(test_glmmtmb@models[[1]], "glmmTMB")
            expect_s3_class(test_glmmtmb_subset@models[[1]], "glmmTMB")

            # Check that glmmTMB has used priors
            expect_no_error(test_glmmtmb@models[[1]]$modelInfo$priors)
            expect_no_error(test_glmmtmb_subset@models[[1]]$modelInfo$priors)


          })

#' Simulate counts from a simple experiment (paired or unpaired) using
#' Poisson or negative binomial distributions.
#'
#' This function is used for internal testing and tutorials.
#'
#'
#'
#'
#' @param n_genes Number of genes to simulate.
#' @param n_samples Number of samples to simulate.
#' @param beta_0 Gene intercepts.
#' @param sigma_0 SD of gene intercepts
#' @param beta_1 Gene slopes.
#' @param sigma_1 SD of gene slopes
#' @param b_0 Cluster-specific intercept. If NULL, non-paired data is simulated.
#' @param clusters Number of clusters. If NULL, non-paired data is simulated.
#' @param sample_sd SD of sample-specific effects.
#' @param overdispersion_min_max Range of overdispersion parameter. If NULL,
#' the Poisson distribution is used to simulate data.
#' @param seed Random seed for reproducibility.
#' @importFrom stats rnorm runif rpois rnbinom
#'
#' @return A list with three elements: `data`, a data frame of simulated
#' counts with one row per gene and one column per sample (first column
#' identifies the gene); `parameters`, a data frame of true gene-wise
#' coefficients and overdispersion values; and `metadata`, a data frame of
#' sample-level covariates (`sample`, `cluster`, `x`).
#'
#' @examples
#' # Simulate n targets from the negative binomial distribution
#' dat <- simcounts(n_genes = 10,
#'                  overdispersion_min_max = c(1, 5))
#'
#' # Simulate n targets from the Poisson distribution
#' dat <- simcounts(n_genes = 10,
#'                  overdispersion_min_max = NULL)
#'
#'
#' # Data are organized in counts
#' dat$data
#' # .. and meta data
#' dat$metadata
#'
#' @export
simcounts <- function(
  n_genes = 10,
  n_samples = 16,
  beta_0 = 1,
  sigma_0 = 0.5,
  beta_1 = 1,
  sigma_1 = 0.5,
  b_0 = 0.5,
  clusters = 8,
  sample_sd = 0.5,
  overdispersion_min_max = c(1, 10),
  seed = 123
) {
  set.seed(seed)
  # Draw gene-specific effects
  beta_0 <- rnorm(n_genes, mean = beta_0, sd = sigma_0)
  beta_1 <- rnorm(n_genes, mean = beta_1, sd = sigma_1)
  # Draw sample-specific effects
  sample_effects <- rnorm(n_samples, mean = 0, sd = sample_sd)
  # Draw cluster-specific effects
  if (!is.null(b_0)) {
    if (is.null(clusters)) {
      stop("Clusters must be provided when b_0 is not NULL.")
    }
    cluster_effects <- rnorm(clusters, mean = 0, sd = b_0)
    cluster_effects <- rep(cluster_effects, n_samples / clusters)
  } else {
    cluster_effects <- rep(0, n_samples)
  }

  # Expression values
  expr_mat <- matrix(0, nrow = n_genes, ncol = n_samples)

  # Design matrix
  x <- rep(c(0, 1), each = n_samples / 2)
  if (!is.null(b_0)) {
    cluster <- rep(1:clusters, length = n_samples)
  }

  # If overdispersion calculate gene-wise overdispersion parameter
  # Draw overdispersion parameter from uniform distribution
  if (!is.null(overdispersion_min_max)) {
    overdispersion <- runif(
      n_genes,
      min = overdispersion_min_max[1],
      max = overdispersion_min_max[2]
    )
  }

  metadata <- data.frame(
    sample = paste0("samp", seq_len(n_samples)),
    cluster = paste0("c", cluster),
    x = x
  )

  for (gene in seq_len(n_genes)) {
    for (sample in seq_len(n_samples)) {
      lambda <- exp(
        beta_0[gene] +
          beta_1[gene] * x[sample] +
          cluster_effects[cluster[sample]] +
          sample_effects[sample]
      )

      if (is.null(overdispersion_min_max)) {
        # Poisson distribution
        expr_mat[gene, sample] <- rpois(1, lambda)
      } else {
        # Negative binomial distribution

        expr_mat[gene, sample] <- rnbinom(
          1,
          size = overdispersion[gene],
          mu = lambda
        )
      }
    }
  }

  data <- data.frame(targetid = paste0("gene", seq_len(n_genes)), expr_mat)

  colnames(data)[-1] <- metadata$sample

  if (is.null(overdispersion_min_max)) {
    # Poisson distribution
    overdispersion <- rep(NA, n_genes)
  }

  parameters <- data.frame(
    targetid = paste0("gene", seq_len(n_genes)),
    beta_0 = beta_0,
    beta_1 = beta_1,
    overdispersion = overdispersion
  )

  # Return list with true parameters,  expression matrix and metadata
  return(list(data = data, parameters = parameters, metadata = metadata))
}

library(testthat)
library(seqwrap)


test_that("simcounts returns the documented structure with defaults", {
  skip_on_cran()
  out <- simcounts()

  # Return should be a list with data, parameters, metadata
  expect_type(out, "list")
  expect_true(all(c("data", "parameters", "metadata") %in% names(out)))

  # data: data.frame with gene rows and sample columns (plus 'targetid')
  expect_s3_class(out$data, "data.frame")
  expect_true("targetid" %in% names(out$data))
  expect_equal(nrow(out$data), 10)           # default n_genes
  expect_equal(ncol(out$data), 1 + 16)       # 1 targetid + default n_samples

  # parameters: one row per gene with beta_0, beta_1, overdispersion
  expect_s3_class(out$parameters, "data.frame")
  expect_true(all(c("targetid", "beta_0", "beta_1", "overdispersion") %in% names(out$parameters)))
  expect_equal(nrow(out$parameters), 10)
  expect_true(is.numeric(out$parameters$beta_0))
  expect_true(is.numeric(out$parameters$beta_1))
  expect_true(is.numeric(out$parameters$overdispersion))
  expect_true(all(is.finite(out$parameters$overdispersion)))
  expect_true(all(out$parameters$overdispersion >= 1 & out$parameters$overdispersion <= 10))

  # metadata: one row per sample with sample, cluster, x
  expect_s3_class(out$metadata, "data.frame")
  expect_true(all(c("sample", "cluster", "x") %in% names(out$metadata)))
  expect_equal(nrow(out$metadata), 16)

  # Column names of data (excluding targetid) should match metadata$sample
  expect_identical(colnames(out$data)[-1], out$metadata$sample)

  # Counts should be non-negative integer-like
  counts <- as.matrix(out$data[,-1])
  expect_true(is.numeric(counts))
  expect_true(all(counts >= 0))
  expect_true(all(counts == floor(counts)))
})

test_that("dimensions match specified n_genes and n_samples", {
  n_genes <- 6
  n_samples <- 12
  clusters <- 6  # must divide n_samples

  out <- simcounts(
    n_genes = n_genes,
    n_samples = n_samples,
    clusters = clusters,
    seed = 1
  )

  # data dims: genes x samples (+ targetid column)
  expect_equal(nrow(out$data), n_genes)
  expect_equal(ncol(out$data), 1 + n_samples)

  # parameters: one row per gene
  expect_equal(nrow(out$parameters), n_genes)

  # metadata: one row per sample
  expect_equal(nrow(out$metadata), n_samples)
})

#' A function to fit models with a chosen fitting algorithm, used in seqwrap.
#' @param fun Name of a fitting function, like glmmTMB::glmmTMB
#' @param arg A list of arguments that can be evaluated by the fitting function
#' @param vars A list of variables to be used in the fitting function
#' @return A fitted model object returned by `fun`.
#' @keywords internal
#' @noRd
fit_fun <- function(fun, arg, vars = NULL) {
  if (!is.null(vars)) {
    # Convert data frame to list if needed
    if (is.data.frame(vars) && nrow(vars) == 1) {
      vars <- as.list(vars)
    }

    # Create an evaluation environment with the variables
    eval_env <- list2env(vars, parent = environment())

    # Helper function to check if x is a formula or unevaluated formula expression
    is_formula_expr <- function(x) {
      # Check for actual formula object
      if (inherits(x, "formula")) return(TRUE)
      # Check for unevaluated formula expression (call with ~ operator)
      if (is.call(x) && length(x) >= 1 && identical(x[[1]], as.symbol("~"))) return(TRUE)
      return(FALSE)
    }

    # Recursive function to replace and evaluate
    replace_and_eval <- function(x, env) {
      if (is_formula_expr(x)) {
        # Don't replace variables inside formulas - they reference data columns
        return(x)
      } else if (is.symbol(x)) {
        # If it's a symbol in our vars, return its value
        var_name <- as.character(x)
        if (var_name %in% names(vars)) {
          return(vars[[var_name]])
        }
        return(x)
      } else if (is.call(x)) {
        # Process each element of the call recursively first
        for (i in seq_along(x)) {
          x[[i]] <- replace_and_eval(x[[i]], env)
        }

        # After replacement, try to evaluate the call
        tryCatch({
          return(eval(x, envir = env))
        }, error = function(e) {
          # If evaluation fails, return the modified call
          return(x)
        })
      } else if (is.pairlist(x)) {
        # Handle pairlists (arguments in alist)
        return(as.pairlist(lapply(x, function(elem) replace_and_eval(elem, env))))
      } else if (is.list(x)) {
        # Handle regular lists
        return(lapply(x, function(elem) replace_and_eval(elem, env)))
      } else {
        return(x)
      }
    }

    # Process the argument list
    arg <- replace_and_eval(arg, eval_env)
  }

  fittedmodel <- do.call(fun, arg)

  return(fittedmodel)
}

#' Special fitting function for lme
#' @param fun The fitting function, nlme::lme
#' @param arg A list of arguments that can be evaluated by the fitting function
#' @param vars A list of variables to be used in the fitting function
#' @return A fitted `lme`/`gls` model object returned by `fun`.
#' @keywords internal
#' @noRd
fit_fun_lme <- function(fun, arg, vars = NULL) {

  # Create an environment for the evaluation
  eval_env <- new.env(parent = environment())

  # If we have target-specific variables, handle replacement and evaluation
  if (!is.null(vars)) {
    # Convert data frame to list if needed
    if (is.data.frame(vars) && nrow(vars) == 1) {
      vars <- as.list(vars)
    }

    # Add variables to the evaluation environment
    for (var_name in names(vars)) {
      eval_env[[var_name]] <- vars[[var_name]]
    }

    # Helper function to check if x is a formula or unevaluated formula expression
    is_formula_expr <- function(x) {
      # Check for actual formula object
      if (inherits(x, "formula")) return(TRUE)
      # Check for unevaluated formula expression (call with ~ operator)
      if (is.call(x) && length(x) >= 1 && identical(x[[1]], as.symbol("~"))) return(TRUE)
      return(FALSE)
    }

    # Recursive function to replace and evaluate
    replace_and_eval <- function(x, env) {
      if (is.symbol(x)) {
        # If it's a symbol in our vars, return its value
        var_name <- as.character(x)
        if (var_name %in% names(vars)) {
          return(vars[[var_name]])
        }
        return(x)
      } else if (is_formula_expr(x)) {
        # Don't replace variables inside formulas - they reference data columns
        return(x)
      } else if (is.call(x)) {
        # Process each element of the call recursively first
        for (i in seq_along(x)) {
          x[[i]] <- replace_and_eval(x[[i]], env)
        }

        # After replacement, try to evaluate the call
        tryCatch({
          return(eval(x, envir = env))
        }, error = function(e) {
          # If evaluation fails, return the modified call
          return(x)
        })
      } else if (is.pairlist(x)) {
        # Handle pairlists (arguments in alist)
        return(as.pairlist(lapply(x, function(elem) replace_and_eval(elem, env))))
      } else if (is.list(x)) {
        # Handle regular lists
        return(lapply(x, function(elem) replace_and_eval(elem, env)))
      } else {
        return(x)
      }
    }

    # Process the argument list
    arg <- replace_and_eval(arg, eval_env)
  }

  # Add data to the environment (after replacement)
  if ("data" %in% names(arg)) {
    data_obj <- arg$data
    for (col_name in names(data_obj)) {
      eval_env[[col_name]] <- data_obj[[col_name]]
    }
  }

  # Call the function with explicit environment
  tryCatch(
    {
      # Make a copy of the arguments
      arg_copy <- arg

      # Ensure formulas have the right environment
      if ("fixed" %in% names(arg_copy) && inherits(arg_copy$fixed, "formula")) {
        environment(arg_copy$fixed) <- eval_env
      }

      if ("random" %in% names(arg_copy)) {
        if (inherits(arg_copy$random, "formula")) {
          environment(arg_copy$random) <- eval_env
        } else if (is.list(arg_copy$random)) {
          for (i in seq_along(arg_copy$random)) {
            if (inherits(arg_copy$random[[i]], "formula")) {
              environment(arg_copy$random[[i]]) <- eval_env
            }
          }
        }
      }

      fittedmodel <- do.call(fun, arg_copy)
      return(fittedmodel)
    },
    error = function(e) {
      # If that fails, try using do.call with the environment parameter
      tryCatch(
        {
          fittedmodel <- do.call(fun, arg, envir = eval_env)
          return(fittedmodel)
        },
        error = function(e2) {
          # If that still fails, try a direct call to lme
          if (
            identical(fun, nlme::lme) ||
            (is.character(fun) && fun == "nlme::lme")
          ) {
            # Extract common parameters
            fixed_form <- arg$fixed
            random_form <- arg$random
            data_obj <- arg$data

            # Try with a minimal set of arguments
            return(nlme::lme(
              fixed = fixed_form,
              random = random_form,
              data = data_obj
            ))
          } else {
            # If all else fails, re-throw the original error
            stop(e)
          }
        }
      )
    }
  )
}

#' Transforma and merge data and fit models. The function is used inside
#' seq_wrapper to combine metadata with target-level data and perform the
#' model fitting.
#' The function is used in a call to pbapply::pblapply.
#' @param x A data frame of target-specific quantities, in seqwrap a list
#' from create_list is used iteratively.
#' @param samp_name Sample names from the upper level function
#' @param metdat Metadata from the upper level function
#' @param add_vars Additional variables to keep from the metadata
#' @param arg_list Arguments from the upper level function
#' @param mt_summary_fun Summary function from the upper level function
#' @param mt_eval_fun Evaluation function from the upper level function
#' @param return_mod Logical, should the models be returned as part of the
#'   results?
#' @param save_mods Logical, should the models be saved?
#' @param mod_path Path to save the models
#' @param ffun the fitting function from the upper level function
#' @return A list with the fitted model (if `return_mod = TRUE`), summary
#'   and evaluation outputs, and any errors/warnings captured during fitting.
#' @importFrom stats as.formula
#' @keywords internal
#' @noRd
seqwrap_mtf <- function(
  x,
  samp_name,
  metdat,
  arg_list,
  add_vars,
  mt_summary_fun,
  mt_eval_fun,
  ffun,
  return_mod,
  save_mods,
  mod_path
) {
  # Extracting the specific target-specific data and transposing
  transposed <- data.frame(t(x[[1]]))
  transposed$temp <- rownames(transposed)
  colnames(transposed)[ncol(transposed)] <- samp_name
  rownames(transposed) <- NULL

  # Merging target-specific data with meta data
  df <- merge(transposed, data.frame(metdat), by = samp_name)

  # Keep only data needed for fitting
  # Improved formula variable extraction for lme compatibility
  parsed <- character(0)

  # Handle standard formula parameter (lm, glm, etc.)
  if ("formula" %in% names(arg_list)) {
    parsed <- c(parsed, all.vars(as.formula(arg_list$formula)))
  }

  # Handle model parameter (various functions)
  if ("model" %in% names(arg_list)) {
    parsed <- c(parsed, all.vars(as.formula(arg_list$model)))
  }

  # Handle fixed/random parameters (for lme, lmer, etc.)
  if ("fixed" %in% names(arg_list)) {
    parsed <- c(parsed, all.vars(as.formula(arg_list$fixed)))
  }

  # Specialized handling for nlme::lme random effects
  if ("random" %in% names(arg_list)) {
    # lme can have random as a formula
    if (inherits(arg_list$random, "formula")) {
      parsed <- c(parsed, all.vars(arg_list$random))
    } # lme can have random as a list of formulas
    else if (is.list(arg_list$random)) {
      for (i in seq_along(arg_list$random)) {
        if (inherits(arg_list$random[[i]], "formula")) {
          parsed <- c(parsed, all.vars(arg_list$random[[i]]))
        } else if (is.call(arg_list$random[[i]])) {
          # Extract variable names from calls (e.g., pdDiag(~time))
          parsed <- c(parsed, all.vars(arg_list$random[[i]]))
        }
      }
    } # Handle the special case where random is a call (typically for nlme)
    else if (is.call(arg_list$random)) {
      parsed <- c(parsed, all.vars(arg_list$random))
    }
  }

  # Remove duplicates
  parsed <- unique(parsed)

  # Filter for columns that actually exist in the data
  parsed <- parsed[parsed %in% colnames(df)]

  # Keep also additional variables that exists in the meta data data set
  if (!is.null(add_vars)) parsed <- c(parsed, add_vars)

  df <- df[, parsed, drop = FALSE]

  # Remove attributes from the list of arguments
  # (this solves an issue when using glmmTMB)
  # Create final arguments list, handling special case for lme
  arguments_final <- append(arg_list, list(data = df))

  # Determine if we're using lme from nlme
  is_lme <- identical(ffun, nlme::lme) ||
    (is.character(ffun) && ffun == "nlme::lme") ||
    any(grepl("lme$", deparse(ffun)))

  # Handle formula environments to ensure proper evaluation
  for (i in seq_along(arguments_final)) {
    if (inherits(arguments_final[[i]], "formula")) {
      # Set formula environment to current environment for all functions
      # This is needed for glmmTMB with complex random effects (e.g., ||)
      # and for lme compatibility
      environment(arguments_final[[i]]) <- environment()
    } else if (is.list(arguments_final[[i]])) {
      # Handle nested formulas in lists (like in random effects)
      for (j in seq_along(arguments_final[[i]])) {
        if (inherits(arguments_final[[i]][[j]], "formula")) {
          environment(arguments_final[[i]][[j]]) <- environment()
        }
      }
    }
  }

  # Making target-wise data available
  # for the model fitting algorithm.
  if (!is.null(x[[2]])) {
    target.wise <- x[[2]]
    # NEW: Convert single-row data frame to named list
    if (is.data.frame(target.wise) && nrow(target.wise) == 1) {
      target.wise <- as.list(target.wise)
    }


  } else {
    target.wise <- NULL
  }
  # Add warning/errors to outputs
  warn <- NULL
  err <- NULL
  warn_sum <- NULL
  warn_eval <- NULL
  err_sum <- NULL
  err_eval <- NULL

  # Adding null values to model outputs
  mod <- NULL
  # Adding null values to outputs from summaries and evaluations
  mod_sum <- NULL
  mod_eval <- NULL

  ## Fit the model with improved error handling
  tryCatch(
    {
      if (is_lme) {
        # For lme, use a more cautious approach
        mod <- fit_fun_lme(ffun, arguments_final, target.wise)
      } else {
        # For other functions, use the original approach
        mod <- fit_fun(ffun, arguments_final, vars = target.wise)
      }
    },
    warning = function(w) warn <<- w,
    error = function(e) err <<- e
  )

  ## Do summarize function if it exists
  if (!is.null(mt_summary_fun)) {
    tryCatch(
      mod_sum <- do.call("mt_summary_fun", list(mod)),
      warning = function(w) warn_sum <<- w,
      error = function(e) err_sum <<- e
    )
  }

  ## Do evaluation function if it exists
  if (!is.null(mt_eval_fun)) {
    tryCatch(
      mod_eval <- do.call("mt_eval_fun", list(mod)),
      warning = function(w) warn_eval <<- w,
      error = function(e) err_eval <<- e
    )
  }

  # Save the model if requested
  if (save_mods) saveRDS(mod, file = paste0(mod_path, "/", names(x), ".rds"))

  # Return the model if requested
  if (return_mod) {
    return(list(
      model = mod,
      summaries = mod_sum,
      evaluation = mod_eval,
      warn = warn,
      err = err,
      warn_sum = warn_sum,
      warn_eval = warn_eval,
      err_sum = err_sum,
      err_eval = err_eval
    ))
  } else {
    return(list(
      summaries = mod_sum,
      evaluation = mod_eval,
      warn = warn,
      err = err,
      warn_sum = warn_sum,
      warn_eval = warn_eval,
      err_sum = err_sum,
      err_eval = err_eval
    ))
  }
}

#' Data helper function.
#'
#' The function takes a named list of data frames
#' and combines them into target-wise data frames
#' for use in subsequent modelling.
#'
#' If the input is a data frame, the function returns a
#' a list of data frames with a single row (y in subsequent modelling)
#'
#' @param dat Sample/target-wise data, either a data frame or a named list.
#' @param targetdat, target-wise data, e.g. targe-wise dispersion
#' @returns A list of lists, each element of the list has a target specific
#' data frame over all samples and a target-specific data frame with values
#' specific to target.
#' @keywords internal
#' @noRd
data_helper <- function(dat, targetdat = NULL, rownames = FALSE) {

  # Convert tibble to data.frame. This assumes that data comes in
  # list or data frame, not nested in tibbles.
  if (any(c("tbl_df", "tbl") %in% class(dat))) dat <- data.frame(dat)

  # Check if input is a list or a data frame
  if (!is.list(dat) && !is.data.frame(dat)) {
    stop("Input must be either a list of data frames or a single data frame")
  }

  # If the input is a data frame
  if (inherits(dat, "data.frame")) {

    if (rownames) {
      cli::cli_inform("Using row names as target identification")

      col_names <- colnames(dat)
      dat[["target_id"]] <- rownames(dat)
      # Move this column to be the first column
      dat <- dat[, c("target_id", col_names)]
      # Reset row names to numeric
      rownames(dat) <- NULL
    }

    # Split the data into a list of data frames for each target
    dfs <- split(dat[, seq_len(ncol(dat))[-1]], dat[, 1])

    # Check for duplicate targets and provide informative error
    dfs <- lapply(names(dfs), function(target_name) {
      x <- dfs[[target_name]]
      if (nrow(x) > 1) {
        stop("Duplicate target found: '", target_name, "' appears ", nrow(x),
             " times in the first column. Each target should appear only once.",
             call. = FALSE)
      }
      rownames(x) <- "y"
      return(x)
    })
    names(dfs) <- names(split(dat[, seq_len(ncol(dat))[-1]], dat[, 1]))


  }

  if (inherits(dat, "list")) {
    if (rownames) {
      cli::cli_inform("Using row names as target identification")
      dat <- lapply(dat, function(x) {
        col_names <- colnames(x)
        x[["target_id"]] <- rownames(x)
        # Move this column to be the first column
        x <- x[, c("target_id", col_names)]
        # Reset row names to numeric
        rownames(x) <- NULL
        return(x)
      })
    }

    # Check if all elements are data frames
    if (!all(sapply(dat, is.data.frame))) {
      stop("All elements in the list must be data frames")
    }

    # Get the names of the input data frames
    df_names <- names(dat)
    if (is.null(df_names)) {
      stop("The list of data frames must be named")
    }

    # Find the number of rows in each data frame
    row_counts <- sapply(dat, nrow)

    # Check if all data frames have the same number of rows
    if (length(unique(row_counts)) != 1) {
      stop("All data frames must have the same number of rows")
    }

    # Check if all data frames have at least one column
    if (any(sapply(dat, ncol) < 1)) {
      stop("All data frames must have at least one column")
    }

    # Create a list to store the new data frames
    dfs <- list()

    # For each row index
    for (i in seq_len(row_counts[1])) {
      # Create a new data frame for this row
      new_df <- data.frame(row.names = df_names)

      # Get the name for this result element (from first column of first data frame)
      first_df <- dat[[1]]
      first_col_name <- colnames(first_df)[1]
      result_name <- as.character(first_df[i, 1])

      # For each original data frame
      for (j in seq_along(dat)) {
        df_name <- df_names[j]
        current_df <- dat[[j]]

        # Extract the row from the current data frame
        row_data <- current_df[i, , drop = FALSE]

        # For each column in the current data frame (skip the first column)
        for (col_name in colnames(current_df)[-1]) {
          # Add the value to the new data frame
          new_df[df_name, col_name] <- row_data[[col_name]]
        }
      }

      # Add the new data frame to the result list with appropriate name
      dfs[[result_name]] <- new_df
    }


  }

  # Include targetwise data if it is not null
  if(!is.null(targetdat)) {
    # Target-wise data, common to all samples
    # Check if input is a data frame, vector, or list
    if (!is.vector(targetdat) && !is.data.frame(targetdat) && !is.list(targetdat)) {
      stop("Target-wise data must be either a vector, a data frame, or a list")
    }

    # If the target-wise data is a vector, adding "target.wise" as heading
    if (is.vector(targetdat)) {
      target.wise.dat <- data.frame(target.wise = targetdat)
      dfs.targetwise <- split(target.wise.dat, seq(nrow(target.wise.dat)))
      names(dfs.targetwise) <- rownames(target.wise.dat)
    }

    # If the target-wise data is a data frame, columns of the data available
    # data frame will be as is
    if (is.data.frame(targetdat)) {
      dfs.targetwise <- split(targetdat, seq(nrow(targetdat)))
      names(dfs.targetwise) <- rownames(targetdat)
    }

    # If the target-wise data is already a list, use it directly
    if (is.list(targetdat) && !is.data.frame(targetdat)) {
      dfs.targetwise <- targetdat
      # Ensure names match the data
      if (is.null(names(dfs.targetwise))) {
        names(dfs.targetwise) <- names(dfs)
      }
    }

    # Combine into a list: target/sample data and target-wise data
    combined.dfs <- mapply(function(x, y) list(x, y),
                           dfs,
                           dfs.targetwise,
                           SIMPLIFY = FALSE)
  }
  # If targetdat is not provided, keep structure of the list anyway
  if (is.null(targetdat)) {
    combined.dfs <- lapply(dfs, function(x) list(x, NULL))
  }

  return(combined.dfs)

}

test_that("data_helper returns a list when given a formatted data frame or list.", {
  local({
    # Set up data frames and lists for data_helper tests

   expect_type(seqwrap:::data_helper(dat = data.frame(
     gene_id = c("aaa", "bbb"),
     sample1 = c(21, 26),
     sample2 = c(11, 22)
   )), "list")

   expect_type(seqwrap:::data_helper(dat = data.frame(
     row.names = c("aaa", "bbb"),
     sample1 = c(21, 26),
     sample2 = c(11, 22)
   ), rownames = TRUE), "list")


    expect_type(seqwrap:::data_helper(dat = list(
        y1 = data.frame(
          row.names = c("aaa", "bbb"),
          sample1 = c(21, 26),
          sample2 = c(11, 22)
          ),
        y2 = data.frame(
          row.names = c("aaa", "bbb"),
          sample1 = c(21, 26),
          sample2 = c(11, 22)
        )), rownames = TRUE), "list")
  })
})

test_that("data_helper throws error on duplicate targets in data frame", {
  # Create a data frame with duplicate target IDs in the first column
  dat_with_duplicates <- data.frame(
    gene_id = c("aaa", "bbb", "aaa", "ccc"),  # "aaa" appears twice
    sample1 = c(21, 26, 30, 15),
    sample2 = c(11, 22, 33, 44)
  )

  # Expect error with informative message about duplicate target
  expect_error(
    seqwrap:::data_helper(dat = dat_with_duplicates),
    "Duplicate target found: 'aaa' appears 2 times in the first column"
  )

  # Test with multiple duplicates
  dat_multiple_duplicates <- data.frame(
    gene_id = c("aaa", "bbb", "aaa", "bbb", "aaa"),
    sample1 = c(1, 2, 3, 4, 5),
    sample2 = c(10, 20, 30, 40, 50)
  )

  # Should error on first duplicate encountered (either "aaa" or "bbb")
  expect_error(
    seqwrap:::data_helper(dat = dat_multiple_duplicates),
    "Duplicate target found"
  )
})

#' Simulate counts from a parallel groups design with three time-points
#'
#'
#' Simulate gene counts from a Negative Binomial distribution conditional
#' on a two-condition, repeated measures experiment with three time points. The
#' function is used for internal testing and tutorials.
#'
#' Fixed effects parameter values are specified as vectors (`beta0`,
#' `conditionB`, `timet2`, etc.) with the common vector length being the
#' number of genes simulated.
#'
#' @details
#' The function simulation of counts from a conditional Negative Binomial
#' distribution:
#' \deqn{y_{i[g]} \sim \text{NB}(\mu_{i[g]}, \phi_{[g]})}
#' \deqn{\text{log}(\mu_{i[g]}) = \beta_0 + \beta1 \text{condition}_B + \\
#'  \beta_2 \text{time}_{t2} + \beta_3 \text{time}_{t3} + \\
#'  \beta_4 \text{condition}_B \times \text{time}_{t2} + \\
#'  \beta_5 \text{condition}_B \times \text{time}_{t3} + \\
#'  \text{offset}(\text{library size}) + \\
#'  \text{b}_0 + \text{b}_1 \times \text{time}_{t2} +
#'  \text{b}_2 \times \text{time}_{t3} }
#' Varying effects are added as:
#' \deqn{\text{b}_l \sim \text{Normal}(0, \sigma_l)}
#' The library size is
#' used as an offset for simulating data. Library sizes are simulated from a
#' log-normal distribution or provided in library_size. In the simulation of
#' counts, the library size is entered as offset after scaling to the median
#' library size. Raw library sizes are included in the meta data.
#' The dispersion parameter can be simulated from a model (lm or loess)
#' provided to the function where the predictor variable should be
#' \eqn{\text{log}(\mu)} and
#' the outcome variable (dispersion) should be \eqn{\text{log}(\phi)}.
#' If no model is provided, values are simulated from
#' a log-normal distribution based on hard coded parameter values
#' from an observed mean-dispersion relationship.
#'
#'
#
#' @param n1 Number clusters in group 1 (control)
#' @param n2 Number of clusters in group 2 (treatment)
#' @param beta0 The intercept parameter on the log scale.
#' @param conditionB Baseline differences between conditions
#' @param timet2 Changes from baseline to time-point 2 in the reference group
#' @param timet3 Changes from baseline to time-point 3 in the reference group
#' @param conditionB_timet2 Difference from reference group in changes from
#' baseline to time-point 2
#' @param conditionB_timet3 Difference from reference group in changes from
#' baseline to time-point 3
#' @param b0 Vector of SD of the between cluster variation in intercept
#' @param b1 Vector of SD of the between cluster variation in timet1 effects
#' @param b2 Vector of SD of the between cluster variation in timet2 effects
#' @param phi_model A model (lm or loess) of a dispersion ~ log_mu relationship
#' @param library_size A vector of library sizes with the length (n1 + n2) * 3
#' or NULL if library sizes are to be simulated
#' @param lib_size_mean Mean of the distribution of library sizes.
#' @param lib_size_cv Coefficient of variation for the distribution of library
#' sizes.
#' @param max_prop The maximum count for a single observation as a proportion
#' of the library size.
#' @importFrom stats rnorm runif rnbinom median model.matrix predict sigma
#' @return A list of (1) simulated counts, (2) the eta, and (3) phi values used
#' for simulating data, and (4) the meta data data frame.
#' @examples
#' # Simulate n_genes number of genes
#' n_genes <- 1000
#'
#' dat <- simcounts2(
#'   n1 = 5,
#'   n2 = 5,
#'   beta0 = rnorm(n_genes, mean = 5),
#'   conditionB =rnorm(n_genes),
#'   timet2 = rnorm(n_genes),
#'   timet3 = rnorm(n_genes),
#'   conditionB_timet2 = rnorm(n_genes),
#'   conditionB_timet3 = rnorm(n_genes),
#'   b0 = abs(rnorm(n_genes)),
#'   b1 = abs(rnorm(n_genes)),
#'   b2 = abs(rnorm(n_genes))
#' )
#'
#' # Data are organized in counts
#' dat$counts
#' # .. and meta data
#' dat$metadata
#'
#'
#' @export
simcounts2 <- function(
  n1 = 5,
  n2 = 5,
  beta0 = c(2.3, 3),
  conditionB = c(0.2, 0.1),
  timet2 = c(0, 0),
  timet3 = c(0.5, 0.25),
  conditionB_timet2 = c(0.1, 0.2),
  conditionB_timet3 = c(0.5, 0.6),
  b0 = c(1, 1),
  b1 = c(0, 0),
  b2 = c(0, 0),
  phi_model = NULL,
  library_size = NULL,
  lib_size_mean = 1e6,
  lib_size_cv = 0.3,
  max_prop = 0.02
) {
  # Generate design data
  len_beta0 <- length(beta0)
  len_b1 <- length(conditionB)
  len_b2 <- length(timet2)
  len_b3 <- length(timet3)
  len_b4 <- length(conditionB_timet2)
  len_b5 <- length(conditionB_timet3)

  # Check that all are similar length
  if (
    !all(sapply(
      list(len_b1, len_b2, len_b3, len_b4, len_b5),
      FUN = identical,
      len_beta0
    ))
  ) {
    stop(
      "All parameter values (conditionB, timet2, ...)
        must be of the same length."
    )
  }

  # Generate library sizes
  if (is.null(library_size)) {
    n_samples <- (n1 + n2) * 3 # total samples

    # Simulate realistic library sizes (log-normal distribution)
    sigma_log <- sqrt(log(1 + lib_size_cv^2))
    mu_log <- log(lib_size_mean) - sigma_log^2 / 2

    library_size <- round(exp(rnorm(n_samples, mu_log, sigma_log)))
  } else {
    # Validate library size length
    if (length(library_size) != (n1 + n2) * 3) {
      stop("library_size must have length (n1 + n2) * 3")
    }
  }

  # A function to predict phi from model parameters
  # or use hard-coded values from approximation
  phi_predict <- function(phi_model, log_mu) {
    if (is.null(phi_model)) {
      phi <- 0.9273216 +
        (-0.3868206) * log_mu +
        0.3816686 * log_mu^2 +
        (-0.0549217) * log_mu^3 +
        0.0021181 * log_mu^4
    } else {
      phi <- predict(phi_model, newdata = data.frame(log_mu = log_mu))
      phi # This is on the log scale (noted in details)
    }

    return(phi)
  }

  # A function to extract the sigma from a phi-prediction model
  phi_sigma <- function(phi_model) {
    if (is.null(phi_model)) {
      phi_sigma <- 4.25
    }

    if (inherits(phi_model, "loess")) {
      phi_sigma <- phi_model$s
    }

    if (inherits(phi_model, "lm")) {
      phi_sigma <- sigma(phi_model)
    }

    return(phi_sigma)
  }

  # The coefficient matrix contain all parameter values collected
  # row-wise.
  coef_mat <- matrix(
    c(beta0, conditionB, timet2, timet3, conditionB_timet2, conditionB_timet3),
    ncol = 6
  )
  colnames(coef_mat) <- c(
    "Intercept",
    "conditionB",
    "timet2",
    "timet3",
    "conditionB:timet2",
    "conditionB:timet3"
  )

  # Create the predictor data frame
  design <- rbind(
    expand.grid(
      id = paste0("A", 1:n1),
      time = c("t1", "t2", "t3"),
      condition = c("A")
    ),
    expand.grid(
      id = paste0("B", 1:n2),
      time = c("t1", "t2", "t3"),
      condition = c("B")
    )
  )

  # Adding library sizes to the design
  design$library_size <- library_size

  # Create a model prediction matrix
  mod_mat <- model.matrix(~ condition * time, data = design)

  # Function to simulate one gene
  simulate_one_gene <- function(i) {
    # Combine fixed effects to get the eta
    eta_fixed <- mod_mat %*% coef_mat[i, ]

    # Add library size offset
    offset <- log(library_size / median(library_size))

    # Adding varying effects (non-correlated)
    b_0 <- c(rep(rnorm(n1, 0, b0[i]), 3), rep(rnorm(n2, 0, b0[i]), 3))
    b_1 <- c(
      rep(0, n1),
      rnorm(n1, 0, b1[i]),
      rep(0, n1),
      rep(0, n2),
      rnorm(n2, 0, b1[i]),
      rep(0, n2)
    )
    b_2 <- c(
      rep(0, n1),
      rep(0, n1),
      rnorm(n1, 0, b2[i]),
      rep(0, n2),
      rep(0, n2),
      rnorm(n2, 0, b2[i])
    )

    # Combining all parameters into the linear predictor eta.
    eta_full <- eta_fixed + offset + b_0 + b_1 + b_2

    # Cap the eta_full at max_prop of median library size
    eta_full <- pmin(eta_full, log(median(library_size) * max_prop))

    # Get the phi from predictive model and combine with phi_sigma
    log_phi <- phi_predict(phi_model, log(mean(exp(eta_fixed))))
    log_phi_sd <- phi_sigma(phi_model)

    phi_g <- exp(rnorm(1, log_phi, log_phi_sd))

    # Simulate from the Negative binomial distribution
    counts <- rnbinom(n = length(eta_full), size = phi_g, mu = exp(eta_full))
    names(counts) <- paste0(design$id, "_", design$time)
    names(eta_full) <- paste0(design$id, "_", design$time)

    # Return as list
    list(
      counts = cbind(data.frame(gene = i), t(counts)),
      eta = cbind(data.frame(gene = i), t(eta_full)),
      phi = cbind(data.frame(gene = i), phi_g)
    )
  }

  # Apply simulation to all genes
  results <- lapply(seq_along(beta0), simulate_one_gene)

  # Extract and combine results
  counts <- do.call(rbind, lapply(results, `[[`, "counts"))
  eta <- do.call(rbind, lapply(results, `[[`, "eta"))
  phi <- do.call(rbind, lapply(results, `[[`, "phi"))

  # Add sample ID to design
  design$seq_sample_id <- paste0(design$id, "_", design$time)

  return(list(counts = counts, eta = eta, phi = phi, metadata = design))
}

library(testthat)
library(seqwrap)



test_that("simcounts2 returns the documented structure with defaults", {

    skip_on_cran() # simulation thus no need to be checked by CRAN

  #### Use variables so that we know nrows, n1 n2 etc?
  n1 <- 6
  n2 <- 5

  out <- simcounts2(n1 = n1, n2 = n2)

  # The documented return is a list with (1) counts, (2) eta, (3) phi, (4) meta
  expect_type(out, "list")
  expect_true(all(c("counts", "eta", "phi", "metadata") %in% names(out)))

  # counts: dataframe of non-negative integers; dimensions: genes x samples
  expect_s3_class(out$counts, "data.frame")
  expect_true(all(apply(out$counts[,-1], 2, is.numeric)))
  expect_true(all(apply(out$counts[,-1], 2, is.finite)))
  expect_true(all(out$counts >= 0))


  # allow >=0 and integer-valued
  expect_true(all(as.matrix(out$counts[,-1]) == floor(as.matrix(out$counts[,-1]))))

  # eta: matrix;
  expect_true(is.data.frame(out$eta))
  # Check that eta and counts have the same dimensions
  expect_equal(dim(out$eta), dim(out$counts))
  # Check that all eta values are numeric
  expect_true(all(apply(out$eta[,-1], 2, is.numeric)))
  # Check tha no values are missing or inf
  expect_true(all(apply(out$eta[,-1], 2, is.finite)))


  # phi: numeric vector (second column) in a data frame,
  expect_true(is.numeric(out$phi[,2]))
  expect_true(all(is.finite(out$phi[,2])))
  # Check that number of columns are greater than 0
  expect_gt(length(out$phi), 0)
  # Check number of rows
  expect_gt(nrow(out$phi), 0)


  # meta: data.frame with one row per sample
  expect_s3_class(out$meta, "data.frame")
  # expected samples: (n1 + n2) * 3 from defaults (5+5)*3 = 30
  expect_equal(nrow(out$meta), (n1 + n2) * 3)

  # The number of columns in counts should match number of samples
  expect_equal(ncol(out$counts[,-1]), nrow(out$meta))

  # The number of genes should match length(beta0) (default beta0 length = 2)
  expect_equal(nrow(out$counts), length(c(2.3, 3)))

  # If library sizes are included in metadata, sanity-check them
  if ("library_size" %in% names(out$meta)) {
    expect_true(is.numeric(out$meta$library_size))
    expect_true(all(out$meta$library_size > 0))
    expect_equal(length(out$meta$library_size), nrow(out$meta))
  }
})

#' Dungan et al 2022 counts
#'
#' Raw count data from Dungan et al. 2022 was downloaded from NCBI Gene
#' Expression Omnibus (GEO) and prepared for use in
#' the seqwrap package. Preparation included adding gene symbols to the first
#' column of the count data frame (counts), and sorting relevant variables to
#' metadata data frame (metadata).
#'
#' @format A list with two data frames:
#' \describe{
#'   \item{counts}{Raw gene expression count matrix (genes x samples).
#'     \describe{
#'       \item{genesymbol}{HGNC gene symbol}
#'       \item{OS.. OV..}{Raw counts, one column for each sample containing integer
#'       read counts, named by experimental id.}
#'     }
#'   }
#'   \item{metadata}{Sample-level metadata.
#'     \describe{
#'       \item{seq_sample_id}{Experimental unit id (mouse identifier)}
#'       \item{treatment}{Senolytic or control (Vehicle) treatment}
#'       \item{surgery}{Surgery inducing overload (synergist ablation) or sham surgury}
#'     }
#'   }
#' }
#' @source \url{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE195707}
#' @references Dungan, C. M. et al. Senolytic treatment rescues blunted
#' muscle hypertrophy in old mice. \emph{GeroScience} 44, 1925\enc{–}{-}1940 (2022).
"dungan_counts"

---
title: "Working with lme4 and nlme in seqwrap"
author:
  - name: Chidimma Echebiri
    affiliation:
      - name: University of Inland Norway
        department: Department of Biotechnology, PhD programme in applied ecology and biotechnology
  - name: Daniel Hammarström
    email: daniel.hammarstrom@inn.no
    affiliation: 
      - name: University of Inland Norway
        department: Department of Public Health and Sport Sciences 
vignette: >
  %\VignetteIndexEntry{Working with lme4 and nlme in seqwrap}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| label: setup
#| include: false

required_pkgs <- c("glmmTMB", "lme4", "nlme")
have_all <- all(vapply(
  required_pkgs,
  requireNamespace,
  logical(1),
  quietly = TRUE
))
knitr::opts_chunk$set(eval = have_all)
```

The `seqwrap` function lets the user specify which modeling algorithm to use when modeling data provided as a data frame and metadata describing each sample. Target-specific models are passed to the results together with optional user-specified summary and model-evaluation functions. The function aims not to limit the user to the preferences of specific models. This allows for comparisons between modeling frameworks and flexibility for many types of high-dimensional data. In this vignette, three popular packages are applied to the same underlying data with varying model definitions.


### Fitting count data to a negative binomial model using `glmmTBM`

We will start by simulating data from a simple experiment where count data has been gathered from 25 clusters in two conditions (e.g. treatment and control). 

```{r}
#| label: simulated-data

library(seqwrap)
library(glmmTMB)


# Simulate data
dat <- seqwrap::simcounts(n_genes = 1000, 
                    n_samples = 50, 
                    beta_0 = 2, 
                    sigma_0 = 0.5, 
                    beta_1 = 0.5,, 
                    sigma_1 = 2, 
                    b_0 = 1,
                    clusters = 25,
                    overdispersion_min = c(1, 10))

# Calculate library sizes
dat$metadata$libsize <- log(colSums(dat$data[, -1]) )


```

Using `seqwrap_compose` we can populate a `swcontainer` object with all needed information to fit models. The `swcontainer` objects is basically a list with data, meta data and arguments that gets passed the model algorithm. The `swcontainer` object also contain reference to the desired model function. 

```{r}

swobject <- seqwrap_compose(
  data = dat$data, # A data frame with the first column as the target identifier
  metadata = dat$metadata, # Meta data bout samples, containing predictors
  samplename = "sample", # The name sample id column 
  modelfun = glmmTMB::glmmTMB, # The model fitting function
  # A list of argumnets supplied to the modelfunction
  arguments = list(
    formula = y ~ x + (1 | cluster),
    family = glmmTMB::nbinom1
  ),
  summary_fun = NULL,
  eval_fun = NULL,
  additional_vars = NULL
)

```


The `seqwrap` function will iterate over all targets in the data and fit the model and apply the summary and evaluation function to each model. Notice that we can use `return_models = TRUE` to save model in the output. Below we also use `subset = 1:10` to only do the first 10 targets. This is useful when we want to check if teh summary and evaluation functions work as intended. The `swobject` created above is updated when we supply ne w arguments to `seqwrap`.


```{r}

models <- seqwrap(swobject, 
                  return_models = TRUE,
                  subset = 1:10) 

```

The resulting object can be summarized using `seqwrap_summarise`

```{r}

modelsum <- seqwrap_summarise(models)

```

We have left the arguments `summary_fun` and `eval_fun` as NULL. This means that `seqwrap` will try to combine model parameter estimates and performance using `broom.mixed::`/`broom::tidy` and `DHARMa`. These evaluations are thought of as minimal summaries of the model and might not completly satisfy the needs of the user. From our example above, the fitted parameters from the models are returned in `modelsum$summaries` and model evaluations are returned in `modelsum$evaluations`.

The functions are exported from the seqwrap package and we can inspect them:

```{r}
seqwrap::generic_summary
```



```{r}
seqwrap::generic_evaluation
```

We encourage users to define their own custom functions. 

### Fitting transformed count data to a Gaussian model using `lme4`

Instead of counts, let's say we are interested in analyzing counts per million, expressed as $\text{CPM} = \frac{\text{Count}}{\text{Library size}} \times 10^6$. We will convert our simulated count data to CPM.


```{r}

lib_sizes <- colSums(dat$data[,-1]) / 1e6

cpm <- sweep(dat$data[,-1], MARGIN = 2, 
             STATS = lib_sizes, 
             FUN = "/")

dat$data <- data.frame(cbind(dat$data[,1], cpm))

```

We will create a new `swcontainer` using `lme4::lmer` as our fitting algorithm. 

```{r}

swobject <- seqwrap_compose(
  data = dat$data, # A data frame with the first column as the target identifier
  metadata = dat$metadata, # Meta data bout samples, containing predictors
  samplename = "sample", # The name sample id column 
  modelfun = lme4::lmer, # The model fitting function
  # A list of argumnets supplied to the modelfunction
  arguments = list(
    formula = y ~ x + (1 | cluster)
  ),
  summary_fun = NULL,
  eval_fun = NULL,
  additional_vars = NULL
)

```

Next we can run `seqwrap` to get estimates for our genes, to save some time we will start by using a subset of the simulated genes.

```{r}

models <- seqwrap(swobject, 
                  return_models = TRUE,
                  subset = 1:10) 

```

The generic summary function will provide fitted parameter statistics.

```{r}

modelsum <- seqwrap_summarise(models)

modelsum$summaries

```

### Using `nlme::lme` and `nlme::gls`, need for `additional_vars`

We will now try to fit the same model as previously fitted in `lme4::lmer` using `nlme::lme`. We can use the same `swcontainer` object and update the relevant lines. Notice that `nlme::lme` requires a different syntax for defining the model. Unsurprisingly, the results are the same @fig-complme. 

```{r}

models_lme <- seqwrap(swobject, 
                  modelfun = nlme::lme,
                  
                  arguments = alist(fixed = y ~ x, 
                                    random = ~ 1|cluster), 
                  return_models = TRUE,
                  eval_fun = NULL,
                  subset = 1:10)

modelsum_lme <- seqwrap_summarise(models_lme)


```

```{r}
lme_est <- data.frame(modelsum_lme$summaries[, c(1,4,5, 6,9)])
lmer_est <- data.frame(modelsum$summaries[, c(1,4,5, 6)])
```


```{r}
#| echo: false
#| fig-cap: "Similar estimates between `lme4::lmer` and `nlme::lme` models"
#| label: fig-complme
plot(lme_est[lme_est$term == "x","estimate"], 
     lmer_est[lmer_est$term == "x","estimate"], 
     xlab = "nlme::lme estimate", 
     ylab = "lme4::lmer estimate")
abline(a = 0, b = 1)

```

`seqwrap` will attempt to extract relevant metadata for use in modeling. This is done by extracting variable names from model formulations. In some cases, this can be difficult because we use additional functionality in some packages. Models constructed in `nlme::gls` are an example. 

We can construct a similar model to `nlme::lme` using `nlme::CorCompSymm`, resulting in a model where the correlations between observations within clusters are uniform (see `?nlme::CorCompSymm`). To help `seqwrap` understand that we want to include `cluster` from the metadata in all fitting objects, we need to include it in `additional_vars`.

```{r}

models_gls <- seqwrap(swobject, 
                  modelfun = nlme::gls,
                  arguments = alist(
                  model = y ~ x,
                  correlation = nlme::corCompSymm(form = ~ 1|cluster) ),
                  additional_vars = "cluster",
                  return_models = TRUE,
                  eval_fun = NULL,
                  subset = 1:10)

modelsum_gls <- seqwrap_summarise(models_gls)

```
As evident from @fig-compgls we can see that the two models will produce slightly different p-values.

```{r}


gls_est <- data.frame(modelsum_gls$summaries[, c(1,2,3,4,6)])


```

```{r}
#| echo: false
#| fig-cap: "Similar estimates between `lme4::lmer` and `nlme::lme` models"
#| label: fig-compgls
plot(-log10(lme_est[lme_est$term == "x","p.value"]), 
     -log10(gls_est[gls_est$term == "x","p.value"]), 
     xlab = "nlme::lme p-value", 
     ylab = "nlme::gls p-value")
abline(a = 0, b = 1)

```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simcounts.R
\name{simcounts}
\alias{simcounts}
\title{Simulate counts from a simple experiment (paired or unpaired) using
Poisson or negative binomial distributions.}
\usage{
simcounts(
  n_genes = 10,
  n_samples = 16,
  beta_0 = 1,
  sigma_0 = 0.5,
  beta_1 = 1,
  sigma_1 = 0.5,
  b_0 = 0.5,
  clusters = 8,
  sample_sd = 0.5,
  overdispersion_min_max = c(1, 10),
  seed = 123
)
}
\arguments{
\item{n_genes}{Number of genes to simulate.}

\item{n_samples}{Number of samples to simulate.}

\item{beta_0}{Gene intercepts.}

\item{sigma_0}{SD of gene intercepts}

\item{beta_1}{Gene slopes.}

\item{sigma_1}{SD of gene slopes}

\item{b_0}{Cluster-specific intercept. If NULL, non-paired data is simulated.}

\item{clusters}{Number of clusters. If NULL, non-paired data is simulated.}

\item{sample_sd}{SD of sample-specific effects.}

\item{overdispersion_min_max}{Range of overdispersion parameter. If NULL,
the Poisson distribution is used to simulate data.}

\item{seed}{Random seed for reproducibility.}
}
\value{
A list with three elements: \code{data}, a data frame of simulated
counts with one row per gene and one column per sample (first column
identifies the gene); \code{parameters}, a data frame of true gene-wise
coefficients and overdispersion values; and \code{metadata}, a data frame of
sample-level covariates (\code{sample}, \code{cluster}, \code{x}).
}
\description{
This function is used for internal testing and tutorials.
}
\examples{
# Simulate n targets from the negative binomial distribution
dat <- simcounts(n_genes = 10,
                 overdispersion_min_max = c(1, 5))

# Simulate n targets from the Poisson distribution
dat <- simcounts(n_genes = 10,
                 overdispersion_min_max = NULL)


# Data are organized in counts
dat$data
# .. and meta data
dat$metadata

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic-summaries.R
\name{generic_evaluation}
\alias{generic_evaluation}
\title{Generic evaluation for model fits}
\usage{
generic_evaluation(x)
}
\arguments{
\item{x}{A model fitted in seqwrap.}
}
\value{
A tidy data frame possible to combine using seqwrap_summarise
}
\description{
When no evaluation function is provided to seqwrap (eval_fun), this function
uses DHARMa to simulate residuals and test for uniformity, dispersion and
outliers. These tests could guide troubleshooting and detection of model
misspecification.
}
\examples{
# The generic summary works on model objects
library(seqwrap)

if (requireNamespace("glmmTMB", quietly = TRUE)) {
  library(glmmTMB)

  dat <- simcounts(n_genes = 1000,
                   n_samples = 30,
                   beta_0 = 5,
                   overdispersion_min_max = c(1, 10))

  counts <- dat$data
  metadata <- dat$metadata
  metadata$ln_libsize <- log(colSums(counts[,-1]))

  # Save counts for gene i in the same data frame
  metadata$y <- as.integer(counts[1, -1])

  m <- glmmTMB(y ~ as.factor(x) + offset(ln_libsize),
               data = metadata,
               family = nbinom2)

  generic_summary(m)
  generic_evaluation(m)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqwrap.R
\name{seqwrap}
\alias{seqwrap}
\title{A flexible upper-level wrapper for iterative modelling using any available
fitting algorithm}
\usage{
seqwrap(
  y = NULL,
  modelfun = NULL,
  arguments = NULL,
  data = NULL,
  metadata = NULL,
  samplename = NULL,
  additional_vars = NULL,
  summary_fun = NULL,
  eval_fun = NULL,
  exported = list(),
  return_models = TRUE,
  save_models = FALSE,
  model_path = NULL,
  subset = NULL,
  cores = 1,
  verbose = TRUE
)
}
\arguments{
\item{y, }{An swcontainer object created with seqwrap_compose, a named list
or a DGEList object.}

\item{modelfun}{A model fitting function like stats::lm,
glmmTMB::glmmTMB or lme4::lmer}

\item{arguments}{An alist or list of arguments to be passed to the fitting
function, this should not contain data. Note that the formula must have
y as the dependent variable.}

\item{data}{A data frame or a list of data frames with targets (e.g. genes,
transcripts) as rows and sample names as columns.
If rownames = FALSE (default), each data frame should have target
identifications as the first column in the data frame(s). If rownames = TRUE
row names will be converted to target identifications. If data is provided as
a list, each element of the list should be named. The corresponding names
be available as variables for the fitting function.}

\item{metadata}{A data frame with sample names (corresponding to column
names in the target matrix)
and design variables.}

\item{samplename}{A character value indicating the variable by which
metadata can merge with the target data. This defaults to "seq_sample_id"
as this is used in the trainomeMetaData package.}

\item{additional_vars}{A vector of additional variables that is contained
in the metadata data set that is needed to fit the model. By default the
metadata is reduced to variables contained in the slots
formula/model/fixed/random in additional arguments.
More variables may be needed for offsets, weights etc.}

\item{summary_fun}{A custom (user-created) function for
evaluating/summarizing models. If NULL, a list of fitted models are returned}

\item{eval_fun}{A custom (user-created) function for model
diagnostics/evaluation. If NULL, no evaluation/diagnostics of models are
returned}

\item{exported}{A list of functions, values etc. to be passed to
summary_fun and eval_fun. This list must contain any functions that
should be used in model summarise or evaluations.}

\item{return_models}{Logical, should models be returned as part of the
output? Save models during development on subsets of the data.
If used on large data sets, this will result in large memory usage.}

\item{save_models}{Logical, should models be saved? Models may be saved
on disk to save working memory.}

\item{model_path}{A character. The path to saved models.}

\item{subset}{A sequence, random samples or integers to indicate which
rows to keep in data. This is useful if you want to test the model in a
subset of targets. If left to the default (NULL), all rows will be used.}

\item{cores}{An integer indicating the number of cores to be used in parallel
computations. If NULL, a sequential for loop is used. If "max", all
available cores are used.}

\item{verbose}{Logical, should the function print diagnostics after checking
the data container?}
}
\value{
A nested list with three upper levels slots: models, a list of
fitted objects; summaries, a list of summaries created from the summary_fun
function; evaluations, a list of diagnostics created from eval_fun.
}
\description{
A flexible upper-level wrapper for iterative modelling using any available
fitting algorithm
}
\details{
This function provides a flexible wrapper to fit, summarize and
evaluate statistical models fitted to high dimensional omics-type data.
Models are fitted and passed to user defined functions to summarize and
evaluate models.
}
\examples{
library(seqwrap)

if (requireNamespace("glmmTMB", quietly = TRUE)) {
library(glmmTMB)

# Simulate n targets
dat <- simcounts(n_genes = 10000)

# Save simulated data as separate objects
counts <- dat$data
metadata <- dat$metadata

# Prepare library size for use as offset
metadata$ln_libsize <- log(colSums(counts[,-1]))


# A mixed effects negative binomial model of RNA-seq counts -----------------

# Populate the seqwrap container
container <- seqwrap_compose(
  modelfun = glmmTMB::glmmTMB,
  arguments = list(
    formula = y ~ x + (1|cluster) + offset(ln_libsize),
    family = glmmTMB::nbinom2()
  ),
  data = counts,
  metadata = metadata,
  samplename = "sample"
)

# Run seqwrap using the container on a subset of targets
results <- seqwrap(container,
                   subset = 1:5,
                   cores = 1)


# Summarise results only contains the subset
summaries <- seqwrap_summarise(results)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqwrap.R
\name{seqwrap_summarise}
\alias{seqwrap_summarise}
\title{Summarise seqwrapResults objects}
\usage{
seqwrap_summarise(
  x,
  summaries = TRUE,
  evaluations = TRUE,
  errors = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{x}{A seqwrapResults object}

\item{summaries}{Logical, should summaries be combined?}

\item{evaluations}{Logical, should evaluations be combined?}

\item{errors}{Logical, should errors be combined?}

\item{verbose, }{Logical should progress be printed? Default TRUE}
}
\value{
A list (invisibly) with up to two data frames: \code{summaries},
combined parameter summaries from each model, and \code{evaluations}, combined
diagnostics from each model. Entries are omitted when the corresponding
slot of \code{x} is empty or the user disables them via the \code{summaries} /
\code{evaluations} arguments.
}
\description{
Summarise seqwrapResults objects
}
\details{
This functions attempts to summarize results from the summary and evaluation
functions applied in each iteration during modelling. The function expects
that the summary and evaluation functions return data frames.
}
\examples{
# Load packages and prepare data for examples -------------------------------

library(seqwrap)

if (requireNamespace("glmmTMB", quietly = TRUE)) {
library(glmmTMB)

# Simulate n targets
dat <- simcounts(n_genes = 10000)

# Save simulated data as separate objects
counts <- dat$data
metadata <- dat$metadata

# Prepare library size for use as offset
metadata$ln_libsize <- log(colSums(counts[,-1]))


# A mixed effects negative binomial model of RNA-seq counts -----------------

# Populate the seqwrap container
container <- seqwrap_compose(
  modelfun = glmmTMB::glmmTMB,
  arguments = list(
    formula = y ~ x + (1|cluster) + offset(ln_libsize),
    family = glmmTMB::nbinom2()
  ),
  data = counts,
  metadata = metadata,
  samplename = "sample"
)

# Run seqwrap using the container on a subset of targets
results <- seqwrap(container,
                   subset = 1:5,
                   cores = 1)


# Summarise results only contains the subset
summaries <- seqwrap_summarise(results)

# Get summaries
summaries$summaries

# Get model evaluations
summaries$evaluations
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simcounts2.R
\name{simcounts2}
\alias{simcounts2}
\title{Simulate counts from a parallel groups design with three time-points}
\usage{
simcounts2(
  n1 = 5,
  n2 = 5,
  beta0 = c(2.3, 3),
  conditionB = c(0.2, 0.1),
  timet2 = c(0, 0),
  timet3 = c(0.5, 0.25),
  conditionB_timet2 = c(0.1, 0.2),
  conditionB_timet3 = c(0.5, 0.6),
  b0 = c(1, 1),
  b1 = c(0, 0),
  b2 = c(0, 0),
  phi_model = NULL,
  library_size = NULL,
  lib_size_mean = 1e+06,
  lib_size_cv = 0.3,
  max_prop = 0.02
)
}
\arguments{
\item{n1}{Number clusters in group 1 (control)}

\item{n2}{Number of clusters in group 2 (treatment)}

\item{beta0}{The intercept parameter on the log scale.}

\item{conditionB}{Baseline differences between conditions}

\item{timet2}{Changes from baseline to time-point 2 in the reference group}

\item{timet3}{Changes from baseline to time-point 3 in the reference group}

\item{conditionB_timet2}{Difference from reference group in changes from
baseline to time-point 2}

\item{conditionB_timet3}{Difference from reference group in changes from
baseline to time-point 3}

\item{b0}{Vector of SD of the between cluster variation in intercept}

\item{b1}{Vector of SD of the between cluster variation in timet1 effects}

\item{b2}{Vector of SD of the between cluster variation in timet2 effects}

\item{phi_model}{A model (lm or loess) of a dispersion ~ log_mu relationship}

\item{library_size}{A vector of library sizes with the length (n1 + n2) * 3
or NULL if library sizes are to be simulated}

\item{lib_size_mean}{Mean of the distribution of library sizes.}

\item{lib_size_cv}{Coefficient of variation for the distribution of library
sizes.}

\item{max_prop}{The maximum count for a single observation as a proportion
of the library size.}
}
\value{
A list of (1) simulated counts, (2) the eta, and (3) phi values used
for simulating data, and (4) the meta data data frame.
}
\description{
Simulate gene counts from a Negative Binomial distribution conditional
on a two-condition, repeated measures experiment with three time points. The
function is used for internal testing and tutorials.
}
\details{
Fixed effects parameter values are specified as vectors (\code{beta0},
\code{conditionB}, \code{timet2}, etc.) with the common vector length being the
number of genes simulated.

The function simulation of counts from a conditional Negative Binomial
distribution:
\deqn{y_{i[g]} \sim \text{NB}(\mu_{i[g]}, \phi_{[g]})}
\deqn{\text{log}(\mu_{i[g]}) = \beta_0 + \beta1 \text{condition}_B + \\
 \beta_2 \text{time}_{t2} + \beta_3 \text{time}_{t3} + \\
 \beta_4 \text{condition}_B \times \text{time}_{t2} + \\
 \beta_5 \text{condition}_B \times \text{time}_{t3} + \\
 \text{offset}(\text{library size}) + \\
 \text{b}_0 + \text{b}_1 \times \text{time}_{t2} +
 \text{b}_2 \times \text{time}_{t3} }
Varying effects are added as:
\deqn{\text{b}_l \sim \text{Normal}(0, \sigma_l)}
The library size is
used as an offset for simulating data. Library sizes are simulated from a
log-normal distribution or provided in library_size. In the simulation of
counts, the library size is entered as offset after scaling to the median
library size. Raw library sizes are included in the meta data.
The dispersion parameter can be simulated from a model (lm or loess)
provided to the function where the predictor variable should be
\eqn{\text{log}(\mu)} and
the outcome variable (dispersion) should be \eqn{\text{log}(\phi)}.
If no model is provided, values are simulated from
a log-normal distribution based on hard coded parameter values
from an observed mean-dispersion relationship.
}
\examples{
# Simulate n_genes number of genes
n_genes <- 1000

dat <- simcounts2(
  n1 = 5,
  n2 = 5,
  beta0 = rnorm(n_genes, mean = 5),
  conditionB =rnorm(n_genes),
  timet2 = rnorm(n_genes),
  timet3 = rnorm(n_genes),
  conditionB_timet2 = rnorm(n_genes),
  conditionB_timet3 = rnorm(n_genes),
  b0 = abs(rnorm(n_genes)),
  b1 = abs(rnorm(n_genes)),
  b2 = abs(rnorm(n_genes))
)

# Data are organized in counts
dat$counts
# .. and meta data
dat$metadata


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqwrap.R
\name{swcontainer}
\alias{swcontainer}
\title{seqwrap container class}
\arguments{
\item{.data}{A list containing initial data (inherited from S7::class_list)}

\item{modelfun}{A function used to fit models}

\item{arguments}{A list of arguments for the fitting function}

\item{data}{A data frame or list of data frames with target data}

\item{rownames}{Logical, should row names be used as target IDs?}

\item{metadata}{A data frame with sample information}

\item{targetdata}{A data frame with target-wise information}

\item{samplename}{Character for sample name identification}

\item{additional_vars}{Character vector of additional variables}

\item{summary_fun}{A function for summarizing models}

\item{eval_fun}{A function for evaluating models}

\item{exported}{A list of objects to export to workers}

\item{model_print}{Character representation of model function}

\item{arguments_print}{Character representation of arguments}
}
\value{
An S7 object of class \code{swcontainer} bundling data, metadata,
and the modelling function plus arguments to be consumed by \code{seqwrap()}.
}
\description{
The seqwrap container is used to store and validate data input
to the to seqwrap.
}
\examples{
# swcontainer is the S7 class produced by seqwrap_compose(). End users
# normally build one via seqwrap_compose() rather than calling this
# constructor directly.

library(seqwrap)

dat <- simcounts(n_genes = 5)

container <- seqwrap_compose(
  modelfun = stats::lm,
  arguments = list(formula = y ~ x),
  data = dat$data,
  metadata = dat$metadata,
  samplename = "sample"
)

# The object returned by seqwrap_compose() is a swcontainer instance
S7::S7_inherits(container, swcontainer)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dungan_counts}
\alias{dungan_counts}
\title{Dungan et al 2022 counts}
\format{
A list with two data frames:
\describe{
\item{counts}{Raw gene expression count matrix (genes x samples).
\describe{
\item{genesymbol}{HGNC gene symbol}
\item{OS.. OV..}{Raw counts, one column for each sample containing integer
read counts, named by experimental id.}
}
}
\item{metadata}{Sample-level metadata.
\describe{
\item{seq_sample_id}{Experimental unit id (mouse identifier)}
\item{treatment}{Senolytic or control (Vehicle) treatment}
\item{surgery}{Surgery inducing overload (synergist ablation) or sham surgury}
}
}
}
}
\source{
\url{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE195707}
}
\usage{
dungan_counts
}
\description{
Raw count data from Dungan et al. 2022 was downloaded from NCBI Gene
Expression Omnibus (GEO) and prepared for use in
the seqwrap package. Preparation included adding gene symbols to the first
column of the count data frame (counts), and sorting relevant variables to
metadata data frame (metadata).
}
\references{
Dungan, C. M. et al. Senolytic treatment rescues blunted
muscle hypertrophy in old mice. \emph{GeroScience} 44, 1925\enc{–}{-}1940 (2022).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqwrap.R
\name{seqwrap_compose}
\alias{seqwrap_compose}
\title{Compose a swcontainer object for use in the seqwrap function.}
\usage{
seqwrap_compose(
  x = NULL,
  modelfun,
  arguments,
  data,
  rownames = FALSE,
  metadata,
  targetdata = NULL,
  samplename = "seq_sample_id",
  additional_vars = NULL,
  summary_fun = NULL,
  eval_fun = NULL,
  exported = list(),
  update = list()
)
}
\arguments{
\item{x}{An optional named list or DGEList object (DESeqDataSet not yet
implemented).}

\item{modelfun}{A model fitting function like stats::lm,
glmmTMB::glmmTMB or lme4::lmer}

\item{arguments}{An alist or list of arguments to be passed to the fitting
function, this should not contain data. Note that the formula must have
y as the dependent variable.}

\item{data}{A data frame or a list of data frames with targets (e.g. genes,
transcripts) as rows and sample names as columns.
If rownames = FALSE (default), each data frame should have target
identifications as the first column in the data frame(s). If rownames = TRUE
row names will be converted to target identifications. If data is provided as
a list, each element of the list should be named. The corresponding names
be available as variables for the fitting function.}

\item{rownames}{should row names in data be used as target identifications?
Defaults to FALSE.}

\item{metadata}{A data frame with sample names (corresponding to column
names in the target matrix)
and design variables.}

\item{targetdata}{A data frame or a list with target-wise values
(e.g. dispersion or start values) for each target. This data is made
available for the model fitting function and can be used to specify target
specific data in each iteration of seqwrap. When a data frame is provided
each row corresponds to the target specific value. When a list is provided,
each column of the data frame is available and can be called by name.}

\item{samplename}{A character value indicating the variable by which
metadata can merge with the target data. This defaults to "seq_sample_id"
as this is used in the trainomeMetaData package.}

\item{additional_vars}{A vector of additional variables that is contained
in the metadata data set that is needed to fit the model. By default the
metadata is reduced to variables contained in the slots
formula/model/fixed/random in additional arguments.
More variables may be needed for offsets, weights etc.}

\item{summary_fun}{A custom (user-created) function for
evaluating/summarizing models. If NULL, a list of fitted models are returned}

\item{eval_fun}{A custom (user-created) function for model
diagnostics/evaluation. If NULL, no evaluation/diagnostics of models are
returned}

\item{exported}{A list of functions, values etc. to be passed to
summary_fun and eval_fun. This list must contain any functions that
should be used in model summarise or evaluations.}

\item{update}{A list of named parameters to update a swcontainer object.}
}
\value{
A swcontainer object for direct use in seqwrap.
}
\description{
This function makes it possible to compose and  run checks on combined
data sets (meta data and target data) and fitting functions to avoid issues
in iterative modelling. See examples and vignettes for details.
}
\examples{
# Load packages and prepare data for examples -------------------------------

library(seqwrap)

if (requireNamespace("glmmTMB", quietly = TRUE)) {
library(glmmTMB)

dat <- simcounts2()

# Save simulated data as separate objects
counts <- dat$counts
metadata <- dat$metadata

# Prepare library size for use as offset
metadata$ln_libsize <- log(metadata$library_size)


# A mixed effects negative binomial model of RNA-seq counts -----------------

# Populate the seqwrap container
container <- seqwrap_compose(
  modelfun = glmmTMB::glmmTMB,
  arguments = list(
    formula = y ~ time * condition + (1|id) + offset(ln_libsize),
    family = glmmTMB::nbinom2()
  ),
  data = counts,
  metadata = metadata,
  samplename = "seq_sample_id"
)

# Run seqwrap using the container
results <- seqwrap(container,
                   cores = 1)




# Summarise results
summaries <- seqwrap_summarise(results)


# Including target-specific data --------------------------------------------

# Target specific data can be supplied to seqwrap_compose to enable, e.g.,
# the use of priors for empirical Bayes shrinkage. In this example we are
# setting a dummy-prior on the `condition` parameter and target-specific
# prior on the dispersion parameter.

fixef_priors <- data.frame(
  prior = "normal(0, 1)",
  class = "fixef",
  coef = "conditionB"
)

dips_priors <- data.frame(
  prior = c(
    "normal(0.5, 0.25)",
    "normal(1, 0.25)"),
  class = "fixef_disp",
  coef = 1
)


# Combine information in target-specific list
prior_list <- list()
for(i in 1:2) {

  prior_list[[i]] <- rbind(
    fixef_priors,
    dips_priors[i,]
  )


}



container <- seqwrap_compose(
  modelfun = glmmTMB::glmmTMB,
  # NOTE: The use of `alist` prevents evaluation of list components
  arguments = alist(
    formula = y ~ time * condition + (1|id) + offset(ln_libsize),
    family = glmmTMB::nbinom2(),
    priors = data.frame(
      prior = prior,
      class = class,
      coef = coef
    )
  ),
  data = counts,
  metadata = metadata,
  targetdata = prior_list,
  samplename = "seq_sample_id"
)

# Run seqwrap using the container
results_prior <- seqwrap(container,
                         # Return models to confirm use of prior information
                         # The use of prior information is not recommended
                         return_models = TRUE,
                         cores = 1)

# Confirm prior information
summary(results_prior@models[[1]])
# Compare to naive model
summary(results@models[[1]])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqwrap.R
\name{print.seqwrapResults}
\alias{print.seqwrapResults}
\title{Print method for objects of class seqwrapResults}
\arguments{
\item{x}{A seqwrapResults object}

\item{...}{Currently unused; included for compatibility with the
\code{print} generic.}
}
\value{
Invisibly returns the object
}
\description{
Invoking the print method on seqwrapResults gives a summary
of the fitted objects.
}
\examples{
# Load packages and prepare data for examples --------------------------------

library(seqwrap)

if (requireNamespace("glmmTMB", quietly = TRUE)) {
library(glmmTMB)

dat <- simcounts2()

# Save simulated data as separate objects
counts <- dat$counts
metadata <- dat$metadata

# Prepare library size for use as offset
metadata$ln_libsize <- log(metadata$library_size)


# A mixed effects negative binomial model of RNA-seq counts ------------------

# Populate the seqwrap container
container <- seqwrap_compose(
  modelfun = glmmTMB::glmmTMB,
  arguments = list(
    formula = y ~ time * condition + (1|id) + offset(ln_libsize),
    family = glmmTMB::nbinom2()
  ),
  data = counts,
  metadata = metadata,
  samplename = "seq_sample_id"
)

# Run seqwrap using the container
results <- seqwrap(container,
                   cores = 1)
print(results)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqwrap.R
\name{seqwrapResults}
\alias{seqwrapResults}
\title{seqwrapResults class}
\arguments{
\item{.data}{A list containing initial data (inherited from S7::class_list)}

\item{models}{A list of fitted model objects}

\item{summaries}{A list of model summaries}

\item{evaluations}{A list of model evaluations}

\item{errors}{A data frame of errors and warnings. defaults to an empty dataframe}

\item{n}{Number of samples}

\item{k}{Number of targets}

\item{call_arguments}{Character string of function arguments used}

\item{call_engine}{Character string of modeling engine used}

\item{elapsed_time}{An proc.time() object giving the time needed to complete
iterative model fitting.}
}
\value{
An S7 object of class \code{seqwrap_results} storing fitted models,
summaries, evaluations, and diagnostic information from a \code{seqwrap()} run.
}
\description{
seqwrapResults constructor function.
}
\examples{
# seqwrapResults is the S7 class returned by seqwrap(). End users do
# not normally call the constructor directly -- it is invoked
# internally once iterative model fitting has finished.

library(seqwrap)

dat <- simcounts(n_genes = 5)

container <- seqwrap_compose(
  modelfun = stats::lm,
  arguments = list(formula = y ~ x),
  data = dat$data,
  metadata = dat$metadata,
  samplename = "sample"
)

results <- seqwrap(container, subset = 1:2, cores = 1)

# The object returned by seqwrap() is a seqwrapResults instance
S7::S7_inherits(results, seqwrapResults)

}
